# -*- test-case-name: axiom.test.historic.test_storeID -*-

from axiom.item import Item
from axiom.attributes import text
from axiom.test.historic.stubloader import saveStub


class Dummy(Item):
    typeName = 'axiom_storeid_dummy'
    schemaVersion = 1

    attribute = text(doc='text', allowNone=False)



class Dummy2(Item):
    typeName = 'axiom_storeid_dummy2'
    schemaVersion = 1

    attribute = text(doc='text', allowNone=False)



def createDatabase(s):
    """
    Populate the given Store with some Dummy items.
    """
    Dummy(store=s, attribute=u'one')
    Dummy(store=s, attribute=u'two')
    i = Dummy(store=s, attribute=u'three')
    Dummy2(store=s, attribute=u'four')
    # Work around https://github.com/twisted/axiom/issues/86
    i.checkpoint()
    i.deleteFromStore()



if __name__ == '__main__':
    saveStub(createDatabase, 0x1240846306fcda3289550cdf9515b2c7111d2bac)




import sys
import readline # Imported for its side-effects
import traceback
from pprint import pprint

from axiom._pysqlite2 import Connection

con = Connection.fromDatabaseName(sys.argv[1])
cur = con.cursor()

while True:
    try:
        cur.execute(raw_input("SQL> "))
        results = list(cur)
        if results:
            pprint(results)
    except:
        traceback.print_exc()



# -*- test-case-name: axiom.test -*-

__metaclass__ = type

import gc
from zope.interface import implements, implementer, Interface

from inspect import getabsfile
from weakref import WeakValueDictionary

from twisted.python import log
from twisted.python.reflect import qual, namedAny
from twisted.python.util import mergeFunctionMetadata
from twisted.application.service import (
    IService, IServiceCollection, MultiService)

from axiom import slotmachine, _schema, iaxiom
from axiom.errors import ChangeRejected, DeletionDisallowed
from axiom.iaxiom import IColumn, IPowerupIndirector

from axiom.attributes import (
    SQLAttribute, _ComparisonOperatorMuxer, _MatchingOperationMuxer,
    _OrderingMixin, _ContainableMixin, Comparable, compare, inmemory,
    reference, text, integer, AND, _cascadingDeletes, _disallows)

_typeNameToMostRecentClass = WeakValueDictionary()

def normalize(qualName):
    """
    Turn a fully-qualified Python name into a string usable as part of a
    table name.
    """
    return qualName.lower().replace('.', '_')

class NoInheritance(RuntimeError):
    """
    Inheritance is as-yet unsupported by XAtop.
    """

class NotInStore(RuntimeError):
    """
    """

class CantInstantiateItem(RuntimeError):
    """You can't instantiate Item directly.  Make a subclass.
    """

class MetaItem(slotmachine.SchemaMetaMachine):
    """Simple metaclass for Item that adds Item (and its subclasses) to
    _typeNameToMostRecentClass mapping.
    """

    def __new__(meta, name, bases, dictionary):
        T = slotmachine.SchemaMetaMachine.__new__(meta, name, bases, dictionary)
        if T.__name__ == 'Item' and T.__module__ == __name__:
            return T
        T.__already_inherited__ += 1
        if T.__already_inherited__ >= 2:
            raise NoInheritance("already inherited from item once: "
                                "in-database inheritance not yet supported")
        if T.typeName is None:
            T.typeName = normalize(qual(T))
        if T.schemaVersion is None:
            T.schemaVersion = 1
        if not T.__legacy__ and T.typeName in _typeNameToMostRecentClass:
            # Let's try not to gc.collect() every time.
            gc.collect()
        if T.typeName in _typeNameToMostRecentClass:
            if T.__legacy__:
                return T
            otherT = _typeNameToMostRecentClass[T.typeName]

            if (otherT.__name__ == T.__name__
                    and getabsfile(T) == getabsfile(otherT)
                    and T.__module__ != otherT.__module__):

                if len(T.__module__) < len(otherT.__module__):
                    relmod = T.__module__
                else:
                    relmod = otherT.__module__

                raise RuntimeError(
                    "Use absolute imports; relative import"
                    " detected for type %r (imported from %r)" % (
                        T.typeName, relmod))

            raise RuntimeError("2 definitions of axiom typename %r: %r %r" % (
                    T.typeName, T, _typeNameToMostRecentClass[T.typeName]))
        _typeNameToMostRecentClass[T.typeName] = T
        return T


    def __cmp__(self, other):
        """
        Ensure stable sorting between Item classes.  This provides determinism
        in SQL generation, which is beneficial for debugging and performance
        purposes.
        """
        if isinstance(other, MetaItem):
            return cmp((self.typeName, self.schemaVersion),
                       (other.typeName, other.schemaVersion))
        return NotImplemented


def noop():
    pass

class _StoreIDComparer(Comparable):
    """
    See Comparable's docstring for the explanation of the requirements of my implementation.
    """
    implements(IColumn)

    def __init__(self, type):
        self.type = type

    def __repr__(self):
        return '<storeID ' + qual(self.type) + '.storeID>'

    def fullyQualifiedName(self):
        # XXX: this is an example of silly redundancy, this really ought to be
        # refactored to work like any other attribute (including being
        # explicitly covered in the schema, which has other good qualities like
        # allowing tables to be VACUUM'd without destroying oid stability and
        # every storeID reference ever. --glyph
        return qual(self.type)+'.storeID'

    # attributes required by ColumnComparer
    def infilter(self, pyval, oself, store):
        return pyval

    def outfilter(self, dbval, oself):
        return dbval

    def getShortColumnName(self, store):
        return store.getShortColumnName(self)

    def getColumnName(self, store):
        return store.getColumnName(self)

    def __get__(self, item, type=None):
        if item is None:
            return self
        else:
            return getattr(item, 'storeID')


class _SpecialStoreIDAttribute(slotmachine.SetOnce):
    """
    Because storeID is special (it's unique, it determines a row's cache
    identity, it's immutable, etc) we don't use a regular SQLAttribute to
    represent it - but it still needs to be compared with other SQL attributes,
    as it is in fact represented by the 'oid' database column.

    I implement set-once semantics to enforce immutability, but delegate
    comparison operations to _StoreIDComparer.
    """
    def __get__(self, oself, type=None):
        if type is not None and oself is None:
            if type._storeIDComparer is None:
                # Reuse the same instance so that the store can use it
                # as a key for various caching, like any other attributes.
                type._storeIDComparer = _StoreIDComparer(type)
            return type._storeIDComparer
        return super(_SpecialStoreIDAttribute, self).__get__(oself, type)


def serviceSpecialCase(item, pups):
    if item._axiom_service is not None:
        return item._axiom_service
    svc = MultiService()
    for subsvc in pups:
        subsvc.setServiceParent(svc)
    item._axiom_service = svc
    return svc




class Empowered(object):
    """
    An object which can have powerups.

    @type store: L{axiom.store.Store}
    @ivar store: Persistence object to which powerups can be added for later
        retrieval.

    @type aggregateInterfaces: C{dict}
    @ivar aggregateInterfaces: Mapping from interface classes to callables
        which will be used to produce corresponding powerups.  The callables
        will be invoked with two arguments, the L{Empowered} for which powerups
        are being loaded and with a list of powerups found in C{store}.  The
        return value is the powerup.  These are used only by the callable
        interface adaption API, not C{powerupsFor}.
    """
    aggregateInterfaces = {
        IService: serviceSpecialCase,
        IServiceCollection: serviceSpecialCase}

    def inMemoryPowerUp(self, powerup, interface):
        """
        Install an arbitrary object as a powerup on an item or store.

        Powerups installed using this method will only exist as long as this
        object remains in memory.  They will also take precedence over powerups
        installed with L{powerUp}.

        @param interface: a zope interface
        """
        self._inMemoryPowerups[interface] = powerup


    def powerUp(self, powerup, interface=None, priority=0):
        """
        Installs a powerup (e.g. plugin) on an item or store.

        Powerups will be returned in an iterator when queried for using the
        'powerupsFor' method.  Normally they will be returned in order of
        installation [this may change in future versions, so please don't
        depend on it].  Higher priorities are returned first.  If you have
        something that should run before "normal" powerups, pass
        POWERUP_BEFORE; if you have something that should run after, pass
        POWERUP_AFTER.  We suggest not depending too heavily on order of
        execution of your powerups, but if finer-grained control is necessary
        you may pass any integer.  Normal (unspecified) priority is zero.

        Powerups will only be installed once on a given item.  If you install a
        powerup for a given interface with priority 1, then again with priority
        30, the powerup will be adjusted to priority 30 but future calls to
        powerupFor will still only return that powerup once.


        If no interface or priority are specified, and the class of the
        powerup has a "powerupInterfaces" attribute (containing
        either a sequence of interfaces, or a sequence of
        (interface, priority) tuples), this object will be powered up
        with the powerup object on those interfaces.

        If no interface or priority are specified and the powerup has
        a "__getPowerupInterfaces__" method, it will be called with
        an iterable of (interface, priority) tuples, collected from the
        "powerupInterfaces" attribute described above. The iterable of
        (interface, priority) tuples it returns will then be
        installed.


        @param powerup: an Item that implements C{interface} (if specified)
        @param interface: a zope interface, or None

        @param priority: An int; preferably either POWERUP_BEFORE,
        POWERUP_AFTER, or unspecified.

        @raise TypeError: raises if interface is IPowerupIndirector You may not
        install a powerup for IPowerupIndirector because that would be
        nonsensical.
        """
        if interface is None:
            for iface, priority in powerup._getPowerupInterfaces():
                self.powerUp(powerup, iface, priority)

        elif interface is IPowerupIndirector:
            raise TypeError(
                "You cannot install a powerup for IPowerupIndirector: " +
                powerup)
        else:
            forc = self.store.findOrCreate(_PowerupConnector,
                                           item=self,
                                           interface=unicode(qual(interface)),
                                           powerup=powerup)
            forc.priority = priority


    def powerDown(self, powerup, interface=None):
        """
        Remove a powerup.

        If no interface is specified, and the type of the object being
        installed has a "powerupInterfaces" attribute (containing
        either a sequence of interfaces, or a sequence of (interface,
        priority) tuples), the target will be powered down with this
        object on those interfaces.

        If this object has a "__getPowerupInterfaces__" method, it
        will be called with an iterable of (interface, priority)
        tuples. The iterable of (interface, priority) tuples it
        returns will then be uninstalled.

        (Note particularly that if powerups are added or removed to the
        collection described above between calls to powerUp and powerDown, more
        powerups or less will be removed than were installed.)
        """
        if interface is None:
            for interface, priority in powerup._getPowerupInterfaces():
                self.powerDown(powerup, interface)
        else:
            for cable in self.store.query(_PowerupConnector,
                                          AND(_PowerupConnector.item == self,
                                              _PowerupConnector.interface == unicode(qual(interface)),
                                              _PowerupConnector.powerup == powerup)):
                cable.deleteFromStore()
                return
            raise ValueError("Not powered up for %r with %r" % (interface,
                                                                powerup))


    def __conform__(self, interface):
        """
        For 'normal' interfaces, returns the first powerup found when doing
        self.powerupsFor(interface).

        Certain interfaces are special - IService from twisted.application
        being the main special case - and will be aggregated according to
        special rules.  The full list of such interfaces is present in the
        'aggregateInterfaces' class attribute.
        """
        if interface is IPowerupIndirector:
            # This would cause an infinite loop, since powerupsFor will try to
            # adapt every powerup to IPowerupIndirector, calling this method.
            return

        pups = self.powerupsFor(interface)
        aggregator = self.aggregateInterfaces.get(interface, None)
        if aggregator is not None:
            return aggregator(self, pups)

        for pup in pups:
            return pup  # return first one, or None if no powerups


    def powerupsFor(self, interface):
        """
        Returns powerups installed using C{powerUp}, in order of descending
        priority.

        Powerups found to have been deleted, either during the course of this
        powerupsFor iteration, during an upgrader, or previously, will not be
        returned.
        """
        inMemoryPowerup = self._inMemoryPowerups.get(interface, None)
        if inMemoryPowerup is not None:
            yield inMemoryPowerup
        if self.store is None:
            return
        name = unicode(qual(interface), 'ascii')
        for cable in self.store.query(
            _PowerupConnector,
            AND(_PowerupConnector.interface == name,
                _PowerupConnector.item == self),
            sort=_PowerupConnector.priority.descending):
            pup = cable.powerup
            if pup is None:
                # this powerup was probably deleted during an upgrader.
                cable.deleteFromStore()
            else:
                indirector = IPowerupIndirector(pup, None)
                if indirector is not None:
                    yield indirector.indirect(interface)
                else:
                    yield pup

    def interfacesFor(self, powerup):
        """
        Return an iterator of the interfaces for which the given powerup is
        installed on this object.

        This is not implemented for in-memory powerups.  It will probably fail
        in an unpredictable, implementation-dependent way if used on one.
        """
        pc = _PowerupConnector
        for iface in self.store.query(pc,
                                      AND(pc.item == self,
                                          pc.powerup == powerup)).getColumn('interface'):
            yield namedAny(iface)


    def _getPowerupInterfaces(self):
        """
        Collect powerup interfaces this object declares that it can be
        installed on.
        """
        powerupInterfaces = getattr(self.__class__, "powerupInterfaces", ())
        pifs = []
        for x in powerupInterfaces:
            if isinstance(x, type(Interface)):
                #just an interface
                pifs.append((x, 0))
            else:
                #an interface and a priority
                pifs.append(x)

        m = getattr(self, "__getPowerupInterfaces__", None)
        if m is not None:
            pifs = m(pifs)
            try:
                pifs = [(i, p) for (i, p) in pifs]
            except ValueError:
                raise ValueError("return value from %r.__getPowerupInterfaces__"
                                 " not an iterable of 2-tuples" % (self,))
        return pifs



def transacted(func):
    """
    Return a callable which will invoke C{func} in a transaction using the
    C{store} attribute of the first parameter passed to it.  Typically this is
    used to create Item methods which are automatically run in a transaction.

    The attributes of the returned callable will resemble those of C{func} as
    closely as L{twisted.python.util.mergeFunctionMetadata} can make them.
    """
    def transactionified(item, *a, **kw):
        return item.store.transact(func, item, *a, **kw)
    return mergeFunctionMetadata(func, transactionified)




def dependentItems(store, tableClass, comparisonFactory):
    """
    Collect all the items that should be deleted when an item or items
    of a particular item type are deleted.

    @param tableClass: An L{Item} subclass.

    @param comparison: A one-argument callable taking an attribute and
    returning an L{iaxiom.IComparison} describing the items to
    collect.

    @return: An iterable of items to delete.
    """
    for cascadingAttr in (_cascadingDeletes.get(tableClass, []) +
                          _cascadingDeletes.get(None, [])):
        for cascadedItem in store.query(cascadingAttr.type,
                                        comparisonFactory(cascadingAttr)):
            yield cascadedItem



def allowDeletion(store, tableClass, comparisonFactory):
    """
    Returns a C{bool} indicating whether deletion of an item or items of a
    particular item type should be allowed to proceed.

    @param tableClass: An L{Item} subclass.

    @param comparison: A one-argument callable taking an attribute and
    returning an L{iaxiom.IComparison} describing the items to
    collect.

    @return: A C{bool} indicating whether deletion should be allowed.
    """
    for cascadingAttr in (_disallows.get(tableClass, []) +
                          _disallows.get(None, [])):
        for cascadedItem in store.query(cascadingAttr.type,
                                        comparisonFactory(cascadingAttr),
                                        limit=1):
            return False

    return True



class Item(Empowered, slotmachine._Strict):
    # Python-Special Attributes
    __metaclass__ = MetaItem

    # Axiom-Special Attributes
    __dirty__ = inmemory()
    __legacy__ = False

    __already_inherited__ = 0

    # Private attributes.
    __store = inmemory()        # underlying reference to the store.

    __everInserted = inmemory() # has this object ever been inserted into the
                                # database?

    __justCreated = inmemory()  # was this object just created, i.e. is there
                                # no committed database representation of it
                                # yet

    __deleting = inmemory()     # has this been marked for deletion at
                                # checkpoint

    __deletingObject = inmemory() # being marked for deletion at checkpoint,
                                  # are we also deleting the central object row
                                  # (True: as in an actual delete) or are we
                                  # simply deleting the data row (False: as in
                                  # part of an upgrade)

    storeID = _SpecialStoreIDAttribute(default=None)
    _storeIDComparer = None
    _axiom_service = inmemory()

    # A mapping from interfaces to in-memory powerups.
    _inMemoryPowerups = inmemory()

    def _currentlyValidAsReferentFor(self, store):
        """
        Is this object currently valid as a reference?  Objects which will be
        deleted in this transaction, or objects which are not in the same store
        are not valid.  See attributes.reference.__get__.
        """
        if store is None:
            # If your store is None, you can refer to whoever you want.  I'm in
            # a store but it doesn't matter that you're not.
            return True
        if self.store is not store:
            return False
        if self.__deletingObject:
            return False
        return True


    def _schemaPrepareInsert(self, store):
        """
        Prepare each attribute in my schema for insertion into a given store,
        either by upgrade or by creation.  This makes sure all references point
        to this store and all relative paths point to this store's files
        directory.
        """
        for name, atr in self.getSchema():
            atr.prepareInsert(self, store)


    def store():
        def get(self):
            return self.__store
        def set(self, store):
            if self.__store is not None:
                raise AttributeError(
                    "Store already set - can't move between stores")

            if store._rejectChanges:
                raise ChangeRejected()

            self._schemaPrepareInsert(store)
            self.__store = store
            oid = self.storeID = self.store.executeSchemaSQL(
                _schema.CREATE_OBJECT, [self.store.getTypeID(type(self))])
            if not self.__legacy__:
                store.objectCache.cache(oid, self)
            if store.autocommit:
                log.msg(interface=iaxiom.IStatEvent,
                        name='database', stat_autocommits=1)

                self.checkpoint()
            else:
                self.touch()
            self.activate()
            self.stored()
        return get, set, """

        A reference to a Store; when set for the first time, inserts this object
        into that store.  Cannot be set twice; once inserted, objects are
        'stuck' to a particular store and must be copied by creating a new
        Item.

        """

    store = property(*store())


    def __repr__(self):
        """
        Return a nice string representation of the Item which contains some
        information about each of its attributes.
        """
        attrs = ", ".join("{n}={v}".format(n=name, v=attr.reprFor(self))
                          for name, attr in sorted(self.getSchema()))
        template = b"{s.__name__}({attrs}, storeID={s.storeID})@{id:#x}"
        return template.format(s=self, attrs=attrs, id=id(self))


    def __subinit__(self, **kw):
        """
        Initializer called regardless of whether this object was created by
        instantiation or loading from the database.
        """
        self._axiom_service = None
        self._inMemoryPowerups = {}
        self.__dirty__ = {}
        to__store = kw.pop('__store', None)
        to__everInserted = kw.pop('__everInserted', False)
        to__justUpgraded = kw.pop('__justUpgraded', False)
        self.__store = to__store
        self.__everInserted = to__everInserted
        self.__deletingObject = False
        self.__deleting = False
        tostore = kw.pop('store',None)

        if not self.__everInserted:
            for (name, attr) in self.getSchema():
                if name not in kw:
                    kw[name] = attr.computeDefault()

        for k, v in kw.iteritems():
            setattr(self, k, v)

        if tostore != None:
            if to__justUpgraded:

                # we can't just set the store, because that allocates an ID.
                # we do still need to do all the attribute prep, make sure
                # references refer to this store, paths are adjusted to point
                # to this store's static offset, etc.

                self._schemaPrepareInsert(tostore)
                self.__store = tostore

                # However, setting the store would normally cache this item as
                # well, so we need to cache it here - unless this is actually a
                # dummy class which isn't real!  In that case don't.
                if not self.__legacy__:
                    tostore.objectCache.cache(self.storeID, self)
                if tostore.autocommit:
                    self.checkpoint()
            else:
                self.store = tostore


    def __init__(self, **kw):
        """
        Create a new Item.  This is called on an item *only* when it is being created
        for the first time, not when it is loaded from the database.  The
        'activate()' hook is called every time an item is loaded from the
        database, as well as the first time that an item is inserted into the
        store.  This will be inside __init__ if you pass a 'store' keyword
        argument to an Item's constructor.

        This takes an arbitrary set of keyword arguments, which will be set as
        attributes on the created item.  Subclasses of Item must honor this
        signature.
        """
        if type(self) is Item:
            raise CantInstantiateItem()
        self.__justCreated = True
        self.__subinit__(**kw)


    def __finalizer__(self):
        return noop


    def existingInStore(cls, store, storeID, attrs):
        """Create and return a new instance from a row from the store."""
        self = cls.__new__(cls)

        self.__justCreated = False
        self.__subinit__(__store=store,
                         storeID=storeID,
                         __everInserted=True)

        schema = self.getSchema()
        assert len(schema) == len(attrs), "invalid number of attributes"
        for data, (name, attr) in zip(attrs, schema):
            attr.loaded(self, data)
        self.activate()
        return self
    existingInStore = classmethod(existingInStore)

    def activate(self):
        """The object was loaded from the store.
        """

    def getSchema(cls):
        """
        return all persistent class attributes
        """
        schema = []
        for name, atr in cls.__attributes__:
            atr = atr.__get__(None, cls)
            if isinstance(atr, SQLAttribute):
                schema.append((name, atr))
        cls.getSchema = staticmethod(lambda schema=schema: schema)
        return schema
    getSchema = classmethod(getSchema)


    def persistentValues(self):
        """
        Return a dictionary of all attributes which will be/have been/are being
        stored in the database.
        """
        return dict((k, getattr(self, k)) for (k, attr) in self.getSchema())


    def touch(self):
        # xxx what
        if self.store is None:
            return
        self.store.changed(self)

    def revert(self):
        if self.__justCreated:
            # The SQL revert has already been taken care of.
            if not self.__legacy__:
                self.store.objectCache.uncache(self.storeID, self)
            return
        self.__dirty__.clear()
        dbattrs = self.store.querySQL(
            self._baseSelectSQL(self.store),
            [self.storeID])[0]

        for data, (name, atr) in zip(dbattrs, self.getSchema()):
            atr.loaded(self, data)

        self.__deleting = False
        self.__deletingObject = False


    def deleted(self):
        """User-definable callback that is invoked when an object is well and truly
        gone from the database; the transaction which deleted it has been
        committed.
        """


    def stored(self):
        """
        User-definable callback that is invoked when an object is placed into a
        Store for the very first time.

        If an Item is created with a store, this will be invoked I{after}
        C{activate}.
        """


    def committed(self):
        """
        Called after the database is brought into a consistent state with this
        object.
        """
        if self.__deleting:
            self.deleted()
            if not self.__legacy__:
                self.store.objectCache.uncache(self.storeID, self)
                self.__store = None
        self.__justCreated = False


    def checkpoint(self):
        """
        Update the database to reflect in-memory changes made to this item; for
        example, to make it show up in store.query() calls where it is now
        valid, but was not the last time it was persisted to the database.

        This is called automatically when in 'autocommit mode' (i.e. not in a
        transaction) and at the end of each transaction for every object that
        has been changed.
        """

        if self.store is None:
            raise NotInStore("You can't checkpoint %r: not in a store" % (self,))


        if self.__deleting:
            if not self.__everInserted:
                # don't issue duplicate SQL and crap; we were created, then
                # destroyed immediately.
                return
            self.store.executeSQL(self._baseDeleteSQL(self.store), [self.storeID])
            # re-using OIDs plays havoc with the cache, and with other things
            # as well.  We need to make sure that we leave a placeholder row at
            # the end of the table.
            if self.__deletingObject:
                # Mark this object as dead.
                self.store.executeSchemaSQL(_schema.CHANGE_TYPE,
                                            [-1, self.storeID])

                # Can't do this any more:
                # self.store.executeSchemaSQL(_schema.DELETE_OBJECT, [self.storeID])

                # TODO: need to measure the performance impact of this, then do
                # it to make sure things are in fact deleted:
                # self.store.executeSchemaSQL(_schema.APP_VACUUM)

            else:
                assert self.__legacy__

            # we're done...
            if self.store.autocommit:
                self.committed()
            return

        if self.__everInserted:
            # case 1: we've been inserted before, either previously in this
            # transaction or we were loaded from the db
            if not self.__dirty__:
                # we might have been checkpointed twice within the same
                # transaction; just don't do anything.
                return
            self.store.executeSQL(*self._updateSQL())
        else:
            # case 2: we are in the middle of creating the object, we've never
            # been inserted into the db before
            schemaAttrs = self.getSchema()

            insertArgs = [self.storeID]
            for (ignoredName, attrObj) in schemaAttrs:
                attrObjDuplicate, attributeValue = self.__dirty__[attrObj.attrname]
                # assert attrObjDuplicate is attrObj
                insertArgs.append(attributeValue)

            # XXX this isn't atomic, gross.
            self.store.executeSQL(self._baseInsertSQL(self.store), insertArgs)
            self.__everInserted = True
        # In case 1, we're dirty but we did an update, synchronizing the
        # database, in case 2, we haven't been created but we issue an insert.
        # In either case, the code in attributes.py sets the attribute *as well
        # as* populating __dirty__, so we clear out dirty and we keep the same
        # value, knowing it's the same as what's in the db.
        self.__dirty__.clear()
        if self.store.autocommit:
            self.committed()

    def upgradeVersion(self, typename, oldversion, newversion, **kw):
        # right now there is only ever one acceptable series of arguments here
        # but it is useful to pass them anyway to make sure the code is
        # functioning as expected
        assert typename == self.typeName, '%r != %r' % (typename, self.typeName)
        assert oldversion == self.schemaVersion
        key = typename, newversion
        T = None
        if key in _legacyTypes:
            T = _legacyTypes[key]
        elif typename in _typeNameToMostRecentClass:
            mostRecent = _typeNameToMostRecentClass[typename]
            if mostRecent.schemaVersion == newversion:
                T = mostRecent
        if T is None:
            raise RuntimeError("don't know about type/version pair %s:%d" % (
                    typename, newversion))
        newTypeID = self.store.getTypeID(T) # call first to make sure the table
                                            # exists for doInsert below

        new = T(store=self.store,
                __justUpgraded=True,
                storeID=self.storeID,
                **kw)

        new.touch()
        new.activate()

        self.store.executeSchemaSQL(_schema.CHANGE_TYPE,
                                    [newTypeID, self.storeID])
        self.deleteFromStore(False)
        return new

    def deleteFromStore(self, deleteObject=True):
        # go grab dependent stuff
        if deleteObject:
            if not allowDeletion(self.store, self.__class__,
                                 lambda attr: attr == self):
                raise DeletionDisallowed(
                    'Cannot delete item; '
                    'has referents with whenDeleted == reference.DISALLOW')

            for dependent in dependentItems(self.store, self.__class__,
                                            lambda attr: attr == self):
                dependent.deleteFromStore()

        self.touch()

        self.__deleting = True
        self.__deletingObject = deleteObject

        if self.store.autocommit:
            self.checkpoint()


    # You may specify schemaVersion and typeName in subclasses
    schemaVersion = None
    typeName = None

    ###### SQL generation ######

    def _baseSelectSQL(cls, st):
        if cls not in st.typeToSelectSQLCache:
            attrs = list(cls.getSchema())
            st.typeToSelectSQLCache[cls] = ' '.join(
                ['SELECT',
                 ', '.join(
                     [st.getShortColumnName(a[1]) for a in attrs]),
                 'FROM',
                 st.getTableName(cls),
                 'WHERE',
                 st.getShortColumnName(cls.storeID),
                 '= ?'])
        return st.typeToSelectSQLCache[cls]

    _baseSelectSQL = classmethod(_baseSelectSQL)

    def _baseInsertSQL(cls, st):
        if cls not in st.typeToInsertSQLCache:
            attrs = list(cls.getSchema())
            qs = ', '.join((['?']*(len(attrs)+1)))
            st.typeToInsertSQLCache[cls] = (
                'INSERT INTO '+
                st.getTableName(cls) + ' (' + ', '.join(
                    [ st.getShortColumnName(cls.storeID) ] +
                    [ st.getShortColumnName(a[1]) for a in attrs]) +
                ') VALUES (' + qs + ')')
        return st.typeToInsertSQLCache[cls]

    _baseInsertSQL = classmethod(_baseInsertSQL)

    def _baseDeleteSQL(cls, st):
        if cls not in st.typeToDeleteSQLCache:
            st.typeToDeleteSQLCache[cls] = ' '.join(['DELETE FROM',
                                                     st.getTableName(cls),
                                                     'WHERE',
                                                     st.getShortColumnName(cls.storeID),
                                                     '= ? '
                                                     ])
        return st.typeToDeleteSQLCache[cls]

    _baseDeleteSQL = classmethod(_baseDeleteSQL)

    def _updateSQL(self):
        # XXX no point in caching for every possible combination of attribute
        # values - probably.  check out how prepared statements are used in
        # python sometime.
        dirty = self.__dirty__.items()
        if not dirty:
            raise RuntimeError("Non-dirty item trying to generate SQL.")
        dirty.sort()
        dirtyColumns = []
        dirtyValues = []
        for dirtyAttrName, (dirtyAttribute, dirtyValue) in dirty:
            dirtyColumns.append(self.store.getShortColumnName(dirtyAttribute))
            dirtyValues.append(dirtyValue)
        stmt = ' '.join([
            'UPDATE', self.store.getTableName(self.__class__), 'SET',
             ', '.join(['%s = ?'] * len(dirty)) %
              tuple(dirtyColumns),
            'WHERE ', self.store.getShortColumnName(type(self).storeID), ' = ?'])
        dirtyValues.append(self.storeID)
        return stmt, dirtyValues


    def getTableName(cls, store):
        """
        Retrieve a string naming the database table associated with this item
        class.
        """
        return store.getTableName(cls)
    getTableName = classmethod(getTableName)


    def getTableAlias(cls, store, currentAliases):
        return None
    getTableAlias = classmethod(getTableAlias)



class _PlaceholderColumn(_ContainableMixin, _ComparisonOperatorMuxer,
                         _MatchingOperationMuxer, _OrderingMixin):
    """
    Wrapper for columns from a L{Placeholder} which provides a fully qualified
    name built with a table alias name instead of the underlying column's real
    table name.
    """
    implements(IColumn)

    def __init__(self, placeholder, column):
        self.type = placeholder
        self.column = column

    def __repr__(self):
        return '<Placeholder %r>' % (self.column,)


    def __get__(self, inst):
        return self.column.__get__(inst)

    def fullyQualifiedName(self):
        return self.column.fullyQualifiedName() + '.<placeholder:%s>' % (
            self.type._placeholderCount,)


    def compare(self, other, op):
        return compare(self, other, op)


    def getShortColumnName(self, store):
        return self.column.getShortColumnName(store)


    def getColumnName(self, store):
        assert self.type._placeholderTableAlias is not None, (
            "Placeholder.getTableAlias() must be called "
            "before Placeholder.attribute.getColumnName()")

        return '%s.%s' % (self.type._placeholderTableAlias,
                          self.column.getShortColumnName(store))

    def infilter(self, pyval, oself, store):
        return self.column.infilter(pyval, oself, store)


    def outfilter(self, dbval, oself):
        return self.column.outfilter(dbval, oself)



_placeholderCount = 0

class Placeholder(object):
    """
    Wrap an existing L{Item} type to provide a different name for it.

    This can be used to join a table against itself which is useful for
    flattening normalized data.  For example, given a schema defined like
    this::

        class Tag(Item):
            taggedObject = reference()
            tagName = text()


        class SomethingElse(Item):
            ...


    It might be useful to construct a query for instances of SomethingElse
    which have been tagged both with C{"foo"} and C{"bar"}::

        t1 = Placeholder(Tag)
        t2 = Placeholder(Tag)
        store.query(SomethingElse, AND(t1.taggedObject == SomethingElse.storeID,
                                       t1.tagName == u"foo",
                                       t2.taggedObject == SomethingElse.storeID,
                                       t2.tagName == u"bar"))
    """
    _placeholderTableAlias = None

    def __init__(self, itemClass):
        global _placeholderCount

        self._placeholderItemClass = itemClass
        self._placeholderCount = _placeholderCount + 1
        _placeholderCount += 1

        self.existingInStore = self._placeholderItemClass.existingInStore


    def __cmp__(self, other):
        """
        Provide a deterministic sort order between Placeholder instances.
        Those instantiated first will compare as less than than instantiated
        later.
        """
        if isinstance(other, Placeholder):
            return cmp(self._placeholderCount, other._placeholderCount)
        return NotImplemented


    def __getattr__(self, name):
        if name == 'storeID' or name in dict(self._placeholderItemClass.getSchema()):
            return _PlaceholderColumn(self, getattr(self._placeholderItemClass, name))
        raise AttributeError(name)


    def getSchema(self):
        # In a MultipleItemQuery, the same table can appear more than
        # once in the "SELECT ..." part of the query, determined by
        # getSchema(). In this case, the correct placeholder names
        # need to be used.
        schema = []
        for (name, atr) in self._placeholderItemClass.getSchema():
            schema.append((
                    name,
                    _PlaceholderColumn(
                        self, getattr(self._placeholderItemClass, name))))
        return schema


    def getTableName(self, store):
        return self._placeholderItemClass.getTableName(store)


    def getTableAlias(self, store, currentAliases):
        if self._placeholderTableAlias is None:
            self._placeholderTableAlias = 'placeholder_' + str(len(currentAliases))
        return self._placeholderTableAlias



_legacyTypes = {}               # map (typeName, schemaVersion) to dummy class

def declareLegacyItem(typeName, schemaVersion, attributes, dummyBases=()):
    """
    Generate a dummy subclass of Item that will have the given attributes,
    and the base Item methods, but no methods of its own.  This is for use
    with upgrading.

    @param typeName: a string, the Axiom TypeName to have attributes for.
    @param schemaVersion: an int, the (old) version of the schema this is a proxy
    for.
    @param attributes: a dict mapping {columnName: attr instance} describing
    the schema of C{typeName} at C{schemaVersion}.

    @param dummyBases: a sequence of 4-tuples of (baseTypeName,
    baseSchemaVersion, baseAttributes, baseBases) representing the dummy bases
    of this legacy class.
    """
    if (typeName, schemaVersion) in _legacyTypes:
        return _legacyTypes[typeName, schemaVersion]
    if dummyBases:
        realBases = [declareLegacyItem(*A) for A in dummyBases]
    else:
        realBases = (Item,)
    attributes = attributes.copy()
    attributes['__module__'] = 'item_dummy'
    attributes['__legacy__'] = True
    attributes['typeName'] = typeName
    attributes['schemaVersion'] = schemaVersion
    result = type(str('DummyItem<%s,%d>' % (typeName, schemaVersion)),
                  realBases,
                  attributes)
    assert result is not None, 'wtf, %r' % (type,)
    _legacyTypes[(typeName, schemaVersion)] = result
    return result



class _PowerupConnector(Item):
    """
    I am a connector between the store and a powerup.
    """
    typeName = 'axiom_powerup_connector'

    powerup = reference()
    item = reference()
    interface = text()
    priority = integer()


POWERUP_BEFORE = 1              # Priority for 'high' priority powerups.
POWERUP_AFTER = -1              # Priority for 'low' priority powerups.



def empowerment(iface, priority=0):
    """
    Class decorator for indicating a powerup's powerup interfaces.

    The class will also be declared as implementing the interface.

    @type iface: L{zope.interface.Interface}
    @param iface: The powerup interface.

    @type priority: int
    @param priority: The priority the powerup will be installed at.
    """
    def _deco(cls):
        cls.powerupInterfaces = (
            tuple(getattr(cls, 'powerupInterfaces', ())) +
            ((iface, priority),))
        implementer(iface)(cls)
        return cls
    return _deco



# -*- test-case-name: axiom.test.test_xatop.ProcessConcurrencyTestCase -*-

# this file is in support of the named test case

import sys

from axiom.store import Store
from twisted.python import filepath

# Open the store so that we get the bad version of the schema
s = Store(filepath.FilePath(sys.argv[1]))

# Alert our parent that we did that
sys.stdout.write("1")
sys.stdout.flush()

# Grab the storeID we are supposed to be reading
sids = sys.stdin.readline()
sid = int(sids)

# load the item we were told to - this should force a schema reload
s.getItemByID(sid)

# let our parent process know that we loaded it successfully
sys.stdout.write("2")
sys.stdout.flush()

# then terminate cleanly



# -*- test-case-name: axiom.test.test_sequence -*-

from axiom.item import Item
from axiom.attributes import reference, integer, AND

class _ListItem(Item):
    typeName = 'list_item'
    schemaVersion = 1

    _index = integer()
    _value = reference()
    _container = reference()

class List(Item):
    typeName = 'list'
    schemaVersion = 1

    length = integer(default=0)

    def __init__(self, *args, **kw):
        super(List, self).__init__(**kw)
        if args:
            self.extend(args[0])

    def _queryListItems(self):
        return self.store.query(_ListItem, _ListItem._container == self)

    def _getListItem(self, index):
        return list(self.store.query(_ListItem,
                                     AND(_ListItem._container == self,
                                         _ListItem._index == index)))[0]

    def _delListItem(self, index, resetIndexes=True):
        for li in self.store.query(_ListItem,
                                   AND(_ListItem._container == self,
                                       _ListItem._index == index)):
            li.deleteFromStore(deleteObject=True)
            break

    def _fixIndex(self, index, truncate=False):
        """
        @param truncate: If true, negative indices which go past the
                         beginning of the list will be evaluated as zero.
                         For example::

                         >>> L = List([1,2,3,4,5])
                         >>> len(L)
                         5
                         >>> L._fixIndex(-9, truncate=True)
                         0
        """
        assert not isinstance(index, slice), 'slices are not supported (yet)'
        if index < 0:
            index += self.length
        if index < 0:
            if not truncate:
                raise IndexError('stored List index out of range')
            else:
                index = 0
        return index

    def __getitem__(self, index):
        index = self._fixIndex(index)
        return self._getListItem(index)._value

    def __setitem__(self, index, value):
        index = self._fixIndex(index)
        self._getListItem(index)._value = value

    def __add__(self, other):
        return list(self) + list(other)
    def __radd__(self, other):
        return list(other) + list(self)

    def __mul__(self, other):
        return list(self) * other
    def __rmul__(self, other):
        return other * list(self)

    def index(self, other, start=0, maximum=None):
        if maximum is None:
            maximum = len(self)
        for pos in range(start, maximum):
            if pos >= len(self):
                break
            if self[pos] == other:
                return pos
        raise ValueError, 'List.index(x): %r not in List' % other

    def __len__(self):
        return self.length

    def __delitem__(self, index):
        assert not isinstance(index, slice), 'slices are not supported (yet)'
        self._getListItem(index).deleteFromStore()
        if index < self.length - 1:
            for item in self.store.query(_ListItem, AND(
                    _ListItem._container == self, _ListItem._index > index)):
                item._index -= 1
        self.length -= 1

    def __contains__(self, value):
        return bool(self.count(value))

    def append(self, value):
        """
        @type value: L{axiom.item.Item}
        @param value: Must be stored in the same L{Store<axiom.store.Store>}
                      as this L{List} instance.
        """
        # XXX: Should List.append(unstoredItem) automatically store the item?
        self.insert(self.length, value)

    def extend(self, other):
        for item in iter(other):
            self.append(item)

    def insert(self, index, value):
        index = self._fixIndex(index, truncate=True)
        # If we do List(length=5).insert(50, x), we don't want
        # x's _ListItem._index to actually be 50.
        index = min(index, self.length)
        # This uses list() in case our contents change halfway through.
        # But does that _really_ work?
        for li in list(self.store.query(_ListItem,
                                        AND(_ListItem._container == self,
                                            _ListItem._index >= index))):
            # XXX: The performance of this operation probably sucks
            # compared to what it would be with an UPDATE.
            li._index += 1
        _ListItem(store=self.store,
                  _value=value,
                  _container=self,
                  _index=index)
        self.length += 1

    def pop(self, index=None):
        if index is None:
            index = self.length - 1
        index = self._fixIndex(index)
        x = self[index]
        del self[index]
        return x

    def remove(self, value):
        del self[self.index(value)]

    def reverse(self):
        # XXX: Also needs to be an atomic action.
        length = 0
        for li in list(self.store.query(_ListItem,
                                        _ListItem._container == self,
                                        sort=_ListItem._index.desc)):
            li._index = length
            length += 1
        self.length = length

    def sort(self, *args):
        # We want to sort by value, not sort by _ListItem.  We could
        # accomplish this by having _ListItem.__cmp__ do something
        # with self._value, but that seemed wrong. This was easier.
        values = [li._value for li in self._queryListItems()]
        values.sort(*args)
        index = 0
        for li in self._queryListItems():
            # XXX: Well, can it?
            assert index < len(values), \
                   '_ListItems were added during a sort (can this happen?)'
            li._index = index
            li._value = values[index]
            index += 1

    def count(self, value):
        return self.store.count(_ListItem, AND(
                _ListItem._container == self, _ListItem._value == value))



from setuptools import setup, find_packages

setup(
    name="Axiom",
    description="An in-process object-relational database",
    url="https://github.com/twisted/axiom",

    maintainer="Divmod, Inc.",
    maintainer_email="support@divmod.org",

    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    install_requires=[
        "Twisted>=13.2.0",
        "Epsilon>=0.7.0"
    ],
    extras_require={
        'test': ['hypothesis[datetime]>=2.0.0,<3.0.0'],
        },
    packages=find_packages() + ['twisted.plugins'],
    scripts=['bin/axiomatic'],
    include_package_data=True,

    license="MIT",
    platforms=["any"],

    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Twisted",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 2 :: Only",
        "Topic :: Database"])




"""
Benchmark the tagsOf method of L{axiom.tags.Catalog}
"""

import time, sys

from epsilon.scripts import benchmark

from axiom import store, item, attributes, tags

N = 30

class TaggedObject(item.Item):
    name = attributes.text()



def main():
    s = store.Store("tags.axiom")
    c = tags.Catalog(store=s)

    objects = []
    def createObjects():
        for x in xrange(N):
            objects.append(TaggedObject(store=s))
    s.transact(createObjects)

    def tagObjects():
        for o in objects:
            for x in xrange(N):
                c.tag(o, unicode(x))
    s.transact(tagObjects)

    def getTags():
        for i in xrange(N):
            for o in objects:
                for t in c.tagsOf(o):
                    pass

    benchmark.start()
    s.transact(getTags)
    benchmark.stop()



if __name__ == '__main__':
    main()



import gc

from twisted.trial.unittest import SynchronousTestCase

from axiom._fincache import FinalizingCache


def noop():
    pass



class Object(object):
    """
    An object which can be stored in a FinalizingCache.
    """
    def __init__(self, name=None):
        self.name = name


    def __finalizer__(self):
        return noop


class FinalizingCacheTests(SynchronousTestCase):
    """
    Tests for L{axiom._fincache.FinalizingCache}.
    """
    def setUp(self):
        self.cache = FinalizingCache()


    def test_nonexistentItem(self):
        """
        Retrieving a nonexistent item from the cache results in L{KeyError}.
        """
        self.assertRaises(KeyError, self.cache.get, 42)


    def test_storeItem(self):
        """
        Storing an object in the cache, then retrieving it, results in the
        exact same object reference.
        """
        o = Object()
        self.cache.cache(42, o)
        self.assertIdentical(o, self.cache.get(42))


    def test_storeItemAndCollect(self):
        """
        An object will be removed from the cache if it is garbage collected.
        """
        o = Object()
        self.cache.cache(42, o)
        del o
        gc.collect()
        self.assertRaises(KeyError, self.cache.get, 42)


    def test_storeOverCollectedItem(self):
        """
        If an item is replaced in the cache, the finalizer from the first item
        does not remove the second item.
        """
        o1 = Object(1)
        self.cache.cache(42, o1)
        gc.disable()
        del o1
        gc.collect()

        o2 = Object(2)
        self.cache.cache(42, o2)
        gc.enable()
        self.assertEqual(2, self.cache.get(42).name)



class SyntheticGCInteractions(SynchronousTestCase):
    """
    Tests for garbage collector interactions that cannot be provoked with
    public APIs.
    """

    def test_collectBeforeCallbacks(self):
        """
        Sometimes, the garbage collector runs and causes weakrefs to expire
        (start returning None when called) before they are finalized (their
        callback is invoked).  When this happens, FinalizingCache keeps track
        of expired cache keys so that future callback invocations will not
        expire any I{new} cache entries with the same key as the
        already-expired ones.

        Presently this only happens on PyPy, and there is no discrete public
        API which causes the garbage collector to call callbacks, only to do
        the collection which will I{eventually} (at some unpredictable, future
        time) call those callbacks, so this is verified against fake weakrefs.
        """
        test = self

        class Expirables(object):
            """
            A colletion of expirable weak references.
            """
            def __init__(self):
                self.references = []

            def ref(self, target, callback=lambda ref: None):
                """
                Create a reference to the given target with the given callback.
                """
                self.references.append(ExpireableRef(target, callback))
                return self.references[-1]

            def expireFor(self, target):
                """
                Expire all weak references that point to C{target} and return
                the number of references expired.
                """
                count = 0
                for ref in self.references:
                    if ref.target is target:
                        ref.expire()
                        count += 1
                return count

            def invokeExpired(self):
                """
                Invoke all expired references.
                """
                for ref in self.references:
                    if ref.expired and not ref.invoked:
                        ref.invoke()

        class ExpireableRef(object):
            """
            A reference that can be expired.
            """
            def __init__(self, target, callback):
                self.target = target
                self.callback = callback
                self.expired = False
                self.invoked = False

            def __call__(self):
                """
                Return the target if we have not yet been expired.
                """
                if not self.expired:
                    return self.target
                else:
                    return None

            def invoke(self):
                """
                Invoke the callback associated with this weakref.
                """
                test.assertEqual(self.expired, True, "Expire before invoking.")
                self.invoked = True
                self.callback(self)

            def expire(self):
                """
                Expire this weakref so it no longer points to its value.
                """
                self.expired = True

        expirables = Expirables()
        cache = FinalizingCache(expirables.ref)
        class identifiable(object):
            def __init__(self, name):
                self.name = name
            def __repr__(self):
                return self.name
            def __finalizer__(self):
                return noop

        a = identifiable("a")
        b = identifiable("b")
        cache.cache(1, a)
        self.assertEqual(expirables.expireFor(a), 1)
        cache.cache(1, b)
        expirables.invokeExpired()
        self.assertIdentical(cache.get(1), b)




import itertools
import time

from testbase import cur

for num in itertools.count():
    cur.execute("select * from foo")
    foovals = cur.fetchall()
    print num, 'I fetched', len(foovals), 'values.', time.ctime()




from epsilon.extime import Time

from axiom.item import Item
from axiom.attributes import text, reference, integer, AND, timestamp

class Tag(Item):
    typeName = 'tag'
    schemaVersion = 1

    name = text(doc="""
    The short string which is being applied as a tag to an Item.
    """)

    created = timestamp(doc="""
    When this tag was applied to the Item to which it applies.
    """)

    object = reference(doc="""
    The Item to which this tag applies.
    """)

    catalog = reference(doc="""
    The L{Catalog} item in which this tag was created.
    """)

    tagger = reference(doc="""
    An optional reference to the Item which is responsible for this tag's
    existence.
    """)



class _TagName(Item):
    """
    Helper class to make Catalog.tagNames very fast.  One of these is created
    for each distinct tag name that is created.  _TagName Items are never
    deleted from the database.
    """
    typeName = 'tagname'

    name = text(doc="""
    The short string which uniquely represents this tag.
    """, indexed=True)

    catalog = reference(doc="""
    The L{Catalog} item in which this tag exists.
    """)



class Catalog(Item):

    typeName = 'tag_catalog'
    schemaVersion = 2

    tagCount = integer(default=0)

    def tag(self, obj, tagName, tagger=None):
        """
        """
        # check to see if that tag exists.  Put the object attribute first,
        # since each object should only have a handful of tags and the object
        # reference is indexed.  As long as this is the case, it doesn't matter
        # whether the name or catalog attributes are indexed because selecting
        # from a small set of results is fast even without an index.
        if self.store.findFirst(Tag,
                                AND(Tag.object == obj,
                                    Tag.name == tagName,
                                    Tag.catalog == self)):
            return

        # if the tag doesn't exist, maybe we need to create a new tagname object
        self.store.findOrCreate(_TagName, name=tagName, catalog=self)

        # Increment only if we are creating a new tag
        self.tagCount += 1
        Tag(store=self.store, object=obj,
            name=tagName, catalog=self,
            created=Time(), tagger=tagger)


    def tagNames(self):
        """
        Return an iterator of unicode strings - the unique tag names which have
        been applied objects in this catalog.
        """
        return self.store.query(_TagName, _TagName.catalog == self).getColumn("name")


    def tagsOf(self, obj):
        """
        Return an iterator of unicode strings - the tag names which apply to
        the given object.
        """
        return self.store.query(
            Tag,
            AND(Tag.catalog == self,
                Tag.object == obj)).getColumn("name")


    def objectsIn(self, tagName):
        return self.store.query(
            Tag,
            AND(Tag.catalog == self,
                Tag.name == tagName)).getColumn("object")



def upgradeCatalog1to2(oldCatalog):
    """
    Create _TagName instances which version 2 of Catalog automatically creates
    for use in determining the tagNames result, but which version 1 of Catalog
    did not create.
    """
    newCatalog = oldCatalog.upgradeVersion('tag_catalog', 1, 2,
                                           tagCount=oldCatalog.tagCount)
    tags = newCatalog.store.query(Tag, Tag.catalog == newCatalog)
    tagNames = tags.getColumn("name").distinct()
    for t in tagNames:
        _TagName(store=newCatalog.store, catalog=newCatalog, name=t)
    return newCatalog

from axiom.upgrade import registerUpgrader
registerUpgrader(upgradeCatalog1to2, 'tag_catalog', 1, 2)



# -*- test-case-name: axiom.test.test_upgrading -*-


from axiom.item import Item
from axiom.attributes import text, integer, reference, inmemory

from axiom.upgrade import registerUpgrader

class ActivateHelper:
    activated = 0
    def activate(self):
        self.activated += 1

class Adventurer(ActivateHelper, Item):
    typeName = 'test_app_player'
    schemaVersion = 2

    name = text()
    activated = inmemory()


class Sword(ActivateHelper, Item):
    typeName = 'test_app_sword'
    schemaVersion = 2

    name = text()
    damagePerHit = integer()
    owner = reference()
    activated = inmemory()


def upgradePlayerAndSword(oldplayer):
    newplayer = oldplayer.upgradeVersion('test_app_player', 1, 2)
    newplayer.name = oldplayer.name

    oldsword = oldplayer.sword

    newsword = oldsword.upgradeVersion('test_app_sword', 1, 2)
    newsword.name = oldsword.name
    newsword.damagePerHit = oldsword.hurtfulness * 2
    newsword.owner = newplayer

    return newplayer, newsword

def player1to2(oldplayer):
    newplayer, newsword = upgradePlayerAndSword(oldplayer)
    return newplayer

def sword1to2(oldsword):
    oldPlayerType = oldsword.store.getOldVersionOf('test_app_player', 1)
    oldplayer = list(oldsword.store.query(oldPlayerType,
                                          oldPlayerType.sword == oldsword))[0]
    newplayer, newsword = upgradePlayerAndSword(oldplayer)
    return newsword


registerUpgrader(sword1to2, 'test_app_sword', 1, 2)
registerUpgrader(player1to2, 'test_app_player', 1, 2)





from zope.interface import Interface, Attribute


class IStatEvent(Interface):
    """
    Marker for a log message that is useful as a statistic.

    Log messages with 'interface' set to this class will be made available to
    external observers.  This is useful for tracking the rate of events such as
    page views.
    """


class IAtomicFile(Interface):
    def __init__(tempname, destdir):
        """Create a new atomic file.

        The file will exist temporarily at C{tempname} and be relocated to
        C{destdir} when it is closed.
        """

    def tell():
        """Return the current offset into the file, in bytes.
        """

    def write(bytes):
        """Write some bytes to this file.
        """

    def close(callback):
        """Close this file.  Move it to its final location.

        @param callback: A no-argument callable which will be invoked
        when this file is ready to be moved to its final location.  It
        must return the segment of the path relative to per-user
        storage of the owner of this file.  Alternatively, a string
        with semantics the same as those previously described for the
        return value of the callable.

        @rtype: C{axiom.store.StoreRelativePath}
        @return: A Deferred which fires with the full path to the file
        when it has been closed, or which fails if there is some error
        closing the file.
        """

    def abort():
        """Give up on this file.  Discard its contents.
        """


class IAxiomaticCommand(Interface):
    """
    Subcommand for 'axiomatic' and 'tell-axiom' command line programs.

    Should subclass twisted.python.usage.Options and provide a command to run.

    '.parent' attribute will be set to an object with a getStore method.
    """

    name = Attribute("""
    """)

    description = Attribute("""
    """)



class IBeneficiary(Interface):
    """
    Interface to adapt to when looking for an appropriate application-level
    object to install powerups on.
    """

    def powerUp(implementor, interface):
        """ Install a powerup on this object.  There is not necessarily any inverse
        powerupsFor on a beneficiary, although there may be; installations may
        be forwarded to a different implementation object, or deferred.
        """

class IPowerupIndirector(Interface):
    """
    Implement this interface if you want to change what is returned from
    powerupsFor for a particular interface.
    """

    def indirect(interface):
        """
        When an item which implements IPowerupIndirector is returned from a
        powerupsFor query, this method will be called on it to give it the
        opportunity to return something other than itself from powerupsFor.

        @param interface: the interface passed to powerupsFor
        @type interface: L{zope.interface.Interface}
        """



class IScheduler(Interface):
    """
    An interface for scheduling tasks.  Quite often the store will be adaptable
    to this; in any Mantissa application, for example; so it is reasonable to
    assume that it is if your application needs to schedule timed events or
    queue tasks.
    """
    def schedule(runnable, when):
        """
        @param runnable: any Item with a 'run' method.

        @param when: a Time instance describing when the runnable's run()
        method will be called.  See extime.Time's documentation for more
        details.
        """



class IQuery(Interface):
    """
    An object that represents a query that can be performed against a database.
    """

    limit = Attribute(
        """
        An integer representing the maximum number of rows to be returned from
        this query, or None, if the query is unlimited.
        """)

    store = Attribute(
        """
        The Axiom store that this query will return results from.
        """)

    def __iter__():
        """
        Retrieve an iterator for the results of this query.

        The query is performed whenever this is called.
        """


    def count():
        """
        Return the number of results in this query.

        NOTE: In most cases, this will have to load all of the rows in this
        query.  It is therefore very slow and should generally be considered
        discouraged.  Call with caution!
        """


    def cloneQuery(limit):
        """
        Create a similar-but-not-identical copy of this query with certain
        attributes changed.

        (Currently this only supports the manipulation of the "limit"
        parameter, but it is the intent that with a richer query-introspection
        interface, this signature could be expanded to support many different
        attributes.)

        @param limit: an integer, representing the maximum number of rows that
        this query should return.

        @return: an L{IQuery} provider with the new limit.
        """



class IColumn(Interface):
    """
    An object that represents a column in the database.
    """
    def getShortColumnName(store):
        """
        @rtype: C{str}
        @return: Just the name of this column.
        """


    def getColumnName(store):
        """
        @rtype: C{str}

        @return: The fully qualified name of this object as a column within the
        database, eg, C{"main_database.some_table.[this_column]"}.
        """

    def fullyQualifiedName():
        """
        @rtype: C{str}

        @return: The fully qualfied name of this object as an attribute in
        Python code, eg, C{myproject.mymodule.MyClass.myAttribute}.  If this
        attribute is represented by an actual Python code object, it will be a
        dot-separated sequence of Python identifiers; otherwise, it will
        contain invalid identifier characters other than '.'.
        """

    def __get__(row):
        """
        @param row: an item that has this column.
        @type row: L{axiom.item.Item}

        @return: The value of the column described by this object, for the given
        row.

        @rtype: depends on the underlying type of the column.
        """



class IOrdering(Interface):
    """
    An object suitable for passing to the 'sort' argument of a query method.
    """
    def orderColumns():
        """
        Return a list of two-tuples of IColumn providers and either C{'ASC'} or
        C{'DESC'} defining this ordering.
        """



class IComparison(Interface):
    """
    An object that represents an in-database comparison.  A predicate that may
    apply to certain items in a store.  Passed as an argument to
    attributes.AND, .OR, and Store.query(...)
    """
    def getInvolvedTables():
        """
        Return a sequence of L{Item} subclasses which are referenced by this
        comparison.  A class may appear at most once.
        """


    def getQuery(store):
        """
        Return an SQL string with ?-style bind parameter syntax thingies.
        """


    def getArgs(store):
        """
        Return a sequence of arguments suitable for use to satisfy the bind
        parameters in the result of L{getQuery}.
        """



class IReliableListener(Interface):
    """
    Receives notification of the existence of Items of a particular type.

    {IReliableListener} providers are given to
    L{IBatchProcessor.addReliableListener} and will then have L{processItem}
    called with items handled by that processor.
    """

    def processItem(item):
        """
        Callback notifying this listener of the existence of the given item.
        """

    def suspend():
        """
        Invoked when notification for this listener is being temporarily
        suspended.

        This should clean up any ephemeral resources held by this listener and
        generally prepare to not do anything for a while.
        """

    def resume():
        """
        Invoked when notification for this listener is being resumed.

        Any actions taken by L{suspend} may be reversed by this method.
        """


LOCAL, REMOTE = range(2)
class IBatchProcessor(Interface):
    def addReliableListener(listener, style=LOCAL):
        """
        Add the given Item to the set which will be notified of Items
        available for processing.

        Note: Each Item is processed synchronously.  Adding too many
        listeners to a single batch processor will cause the L{step}
        method to block while it sends notification to each listener.

        @type listener: L{IReliableListener}
        @param listener: The item to which listened-for items will be passed
        for processing.
        """


    def removeReliableListener(listener):
        """
        Remove a previously added listener.
        """


    def getReliableListeners():
        """
        Return an iterable of the listeners which have been added to
        this batch processor.
        """



class IBatchService(Interface):
    """
    Object which allows minimal communication with L{IReliableListener}
    providers which are running remotely (that is, with the L{REMOTE} style).
    """
    def start():
        """
        Start the remote batch process if it has not yet been started, otherwise
        do nothing.
        """


    def suspend(storeID):
        """
        @type storeID: C{int}
        @param storeID: The storeID of the listener to suspend.

        @rtype: L{twisted.internet.defer.Deferred}
        @return: A Deferred which fires when the listener has been suspended.
        """


    def resume(storeID):
        """
        @type storeID: C{int}
        @param storeID: The storeID of the listener to resume.

        @rtype: L{twisted.internet.defer.Deferred}
        @return: A Deferred which fires when the listener has been resumed.
        """



class IVersion(Interface):
    """
    Object with version information for a package that creates Axiom
    items, most likely a L{twisted.python.versions.Version}. Used to
    track which versions of a package have been used to load a store.
    """
    package = Attribute("""
    Name of a Python package.
    """)
    major = Attribute("""
    Major version number.
    """)
    minor = Attribute("""
    Minor version number.
    """)
    micro = Attribute("""
    Micro version number.
    """)



# -*- test-case-name: axiom.test.test_upgrading.DeletionTest.testPowerups -*-

from axiom.item import Item
from axiom.attributes import integer

class Obsolete(Item):
    """
    This is an obsolete class that will be destroyed in the upcoming version.
    """
    typeName = 'test_upgrading_obsolete'
    nothing = integer()




import operator, random

from twisted.trial.unittest import TestCase

from axiom.iaxiom import IComparison, IColumn
from axiom.store import Store, ItemQuery, MultipleItemQuery
from axiom.item import Item, Placeholder

from axiom import errors
from axiom.attributes import (
    reference, text, bytes, integer, AND, OR, TableOrderComparisonWrapper)

class A(Item):
    schemaVersion = 1
    typeName = 'a'

    reftoc = reference()
    type = text(indexed=True)


class B(Item):
    schemaVersion = 1
    typeName = 'b'

    cref = reference()
    name = text(indexed=True)

class C(Item):
    schemaVersion = 1
    typeName = 'c'

    name = text(indexed=True)

class D(Item):
    schemaVersion = 1
    typeName = 'd'
    id = bytes()
    one = bytes()
    two = bytes()
    three = bytes()
    four = text()

class E(Item):
    schemaVersion = 1
    typeName = 'e'
    name = text()
    transaction = text()
    amount = integer()


class ThingWithCharacterAndByteStrings(Item):
    schemaVersion = 1

    typeName = 'ThingWithCharacterAndByteStrings'

    characterString = text(caseSensitive=True)
    caseInsensitiveCharString = text(caseSensitive=False)

    byteString = bytes()


class BasicQuery(TestCase):

    def test_rightHandStoreIDComparison(self):
        """
        Test that a StoreID column on the right-hand side of an equality test
        results in a TwoAttributeComparison object rather than an
        AttributeValueComparison or anything else that would be wrong.
        """
        s = Store()
        comparison = (A.reftoc == B.storeID)
        self.assertEquals(
            comparison.getQuery(s),
            '(%s.[reftoc] = %s.oid)' % (
                A.getTableName(s),
                B.getTableName(s)))
        self.assertEquals(comparison.getArgs(s), [])


    def test_leftHandStoreIDComparison(self):
        """
        Test that a StoreID column on the left-hand side of an equality test
        results in a TwoAttributeComparison object rather than an
        AttributeValueComparison or anything else that would be wrong.
        """
        s = Store()
        comparison = (B.storeID == A.reftoc)
        self.assertEquals(
            comparison.getQuery(s),
            '(%s.oid = %s.[reftoc])' % (
                B.getTableName(s),
                A.getTableName(s)))
        self.assertEquals(comparison.getArgs(s), [])


    def test_simplestQuery(self):
        """
        Test that an ItemQuery with no comparison, sorting, or limit generates
        the right SQL for that operation.
        """
        s = Store()
        query = ItemQuery(s, A)
        sql, args = query._sqlAndArgs('SELECT', '*')
        self.assertEquals(
            sql,
            'SELECT * FROM %s' % (A.getTableName(s),))
        self.assertEquals(args, [])


    def test_simpleIntegerComparison(self):
        """
        Test that an ItemQuery with a single attribute comparison on an integer
        attribute generates SQL with the right WHERE clause.
        """
        s = Store()
        query = ItemQuery(s, E, E.amount == 0)
        sql, args = query._sqlAndArgs('SELECT', '*')
        self.assertEquals(
            sql,
            'SELECT * FROM %s WHERE (%s.[amount] = ?)' % (
                E.getTableName(s),
                E.getTableName(s)))
        self.assertEquals(args, [0])


    def test_simpleReferenceComparison(self):
        """
        Test that an ItemQuery with a single attribute comparison on a
        reference attribute generates SQL with the right WHERE clause.
        """
        s = Store()
        query = ItemQuery(s, A, A.reftoc == A.storeID)
        sql, args = query._sqlAndArgs('SELECT', '*')
        self.assertEquals(
            sql,
            'SELECT * FROM %s WHERE (%s.[reftoc] = %s.oid)' % (
                A.getTableName(s),
                A.getTableName(s),
                A.getTableName(s)))
        self.assertEquals(args, [])


    def test_reversedReferenceComparison(self):
        """
        Test that an ItemQuery with a single attribute comparison on a
        reference attribute with the storeID part on the left-hand side
        generates SQL with the right WHERE clause.
        """
        s = Store()
        query = ItemQuery(s, A, A.storeID == A.reftoc)
        sql, args = query._sqlAndArgs('SELECT', '*')
        self.assertEquals(
            sql,
            'SELECT * FROM %s WHERE (%s.oid = %s.[reftoc])' % (
                A.getTableName(s),
                A.getTableName(s),
                A.getTableName(s)))
        self.assertEquals(args, [])


    def test_unionComparison(self):
        """
        Test that an ItemQuery with two comparisons joined with an L{AND}
        generates the right WHERE clause.
        """
        s = Store()
        query = ItemQuery(s, A, AND(A.reftoc == B.storeID,
                                    B.cref == C.storeID))
        sql, args = query._sqlAndArgs('SELECT', '*')
        self.assertEquals(
            sql,
            'SELECT * FROM %s, %s, %s '
            'WHERE ((%s.[reftoc] = %s.oid) AND '
                   '(%s.[cref] = %s.oid))' % (
                A.getTableName(s),
                B.getTableName(s),
                C.getTableName(s),
                A.getTableName(s),
                B.getTableName(s),
                B.getTableName(s),
                C.getTableName(s)))
        self.assertEquals(args, [])


    def testBasicQuery(self):
        s = Store()

        def entesten():
            c1 = C(store=s,
                   name=u'yes')

            c2 = C(store=s,
                   name=u'no')

            A(store=s,
              reftoc=c1,
              type=u'testc')

            A(store=s,
              reftoc=c2,
              type=u'testc')

            A(store=s,
              reftoc=c1,
              type=u'testx')

            yesb = B(store=s,
                     cref=c1,
                     name=u'correct')

            B(store=s,
              cref=c2,
              name=u'not correct')

            s.checkpoint()

            q = list(s.query(B,
                             AND(AND(C.name == u'yes',
                                     A.type == u'testc'),
                                 AND(C.storeID == B.cref,
                                     A.reftoc == C.storeID)),
                             ))

            self.assertEquals(q, [yesb])

        s.transact(entesten)
        s.close()


    def testStringQueries(self):
        s = Store()

        def createAndStuff():
            text1 = u'Hello, \u1234 world.'
            text2 = u'ThIs sTrInG iS nOt cAsE sEnSiTIvE.  \u4567'
            bytes1 = '\x00, punk'

            x = ThingWithCharacterAndByteStrings(
                store=s,
                characterString=text1,
                caseInsensitiveCharString=text2,
                byteString=bytes1)

            x.checkpoint()

            q = list(
                s.query(ThingWithCharacterAndByteStrings,
                        ThingWithCharacterAndByteStrings.characterString == text1.lower(),
                        ))
            self.failIf(q, q)

            q = list(
                s.query(ThingWithCharacterAndByteStrings,
                        ThingWithCharacterAndByteStrings.characterString == text1.upper(),
                        ))
            self.failIf(q, q)

            q = list(
                s.query(ThingWithCharacterAndByteStrings,
                        ThingWithCharacterAndByteStrings.characterString == text1,
                        ))

            self.assertEquals(q, [x])

            q = list(
                s.query(ThingWithCharacterAndByteStrings,
                        ThingWithCharacterAndByteStrings.caseInsensitiveCharString == text2,
                        ))

            self.assertEquals(q, [x])

            q = list(
                s.query(ThingWithCharacterAndByteStrings,
                        ThingWithCharacterAndByteStrings.caseInsensitiveCharString == text2.lower(),
                        ))

            self.assertEquals(q, [x])

            q = list(
                s.query(ThingWithCharacterAndByteStrings,
                        ThingWithCharacterAndByteStrings.caseInsensitiveCharString == text2.upper(),
                        ))

            self.assertEquals(q, [x])

            q = list(
                s.query(ThingWithCharacterAndByteStrings,
                        ThingWithCharacterAndByteStrings.byteString == bytes1,
                        ))

            self.assertEquals(q, [x])

            q = list(
                s.query(ThingWithCharacterAndByteStrings,
                        ThingWithCharacterAndByteStrings.byteString == bytes1.upper(),
                        ))

            self.failIf(q, q)

        s.transact(createAndStuff)
        s.close()


    def testAggregateQueries(self):
        s = Store()
        def entesten():
            self.assertEquals(s.query(E).count(), 0)
            self.assertEquals(s.query(E).getColumn("amount").sum(), 0)

            E(store=s, name=u'widgets', amount=37)
            E(store=s, name=u'widgets', amount=63)
            E(store=s, name=u'quatloos', amount=99, transaction=u'yes')
            s.checkpoint()
            q = s.count(E, E.name == u'widgets')
            self.failUnlessEqual(q, 2)
            q = s.sum(E.amount, E.name == u'widgets')
            self.failUnlessEqual(q, 100)
        s.transact(entesten)
        s.close()

    def testAttributeQueries(self):
        s = Store()
        def entesten():
            E(store=s, name=u'b', amount=456)
            E(store=s, name=u'a', amount=123)
            E(store=s, name=u'c', amount=789)
            self.assertEquals(list(s.query(E, sort=E.name.ascending).getColumn("amount")),
                              [123, 456, 789])

        s.transact(entesten)
        s.close()

    def testAttributeQueryCount(self):
        s = Store()
        def entesten():
            E(store=s, name=u'a', amount=123)
            E(store=s, name=u'b', amount=456)
            E(store=s, name=u'c')  # no amount given
            self.assertEquals(s.query(E).getColumn("amount").count(), 2)
        s.transact(entesten)
        s.close()

    def testAttributeQueryDistinct(self):
        s = Store()
        def entesten():
            E(store=s, name=u'a', amount=123)
            E(store=s, name=u'b', amount=789)
            E(store=s, name=u'a', amount=456)
            self.assertEquals(list(s.query(E, sort=E.name.ascending).getColumn("name").distinct()),
                              [u"a", u"b"])
        s.transact(entesten)
        s.close()


    def test_distinctQuerySQLiteBug(self):
        """
        Test for an SQLite bug.

        SQLite versions 3.5.8 and 3.5.9 have a bug that causes incorrect query
        results under certain circumstances:

            - A column with an index
            - SELECT DISTINCT query...
            - ... with an IS NOT NULL comparison on that column.

        Upstream ticket: http://www.sqlite.org/cvstrac/tktview?tn=3236
        """
        s = Store()
        def entesten():
            C(store=s, name=u'a')
            C(store=s, name=u'b')
            q = s.query(C, C.name != None).getColumn('name').distinct()
            self.assertEqual(sorted(q), [u'a', u'b'])
        s.transact(entesten)
        s.close()


    def test_itemQueryDistinct(self):
        """
        Verify that a join which would produce duplicate rows will not produce
        duplicate item results.
        """
        s = Store()
        def entesten():
            theC = C(store=s, name=u'it')
            B(store=s, cref=theC)
            B(store=s, cref=theC)
            B(store=s, cref=theC)
            self.assertEquals(list(s.query(C, B.cref == C.storeID).distinct()),
                              [theC])
        s.transact(entesten)
        s.close()


    def test_itemQueryDistinctCount(self):
        """
        Like L{test_itemQueryDistinct} but for the C{count} method of a
        distinct item query.
        """
        s = Store()
        def entesten():
            for n, name in (2, u'it'), (3, u'indefinite nominative'):
                theC = C(store=s, name=name)
                for i in range(n):
                    B(store=s, cref=theC)
            self.assertEquals(
                s.query(C, B.cref == C.storeID).distinct().count(),
                2)
        s.transact(entesten)
        s.close()


    def test_itemQueryLimitAttribute(self):
        """
        L{ItemQuery} implements LI{IQuery} and should provide a 'limit'
        attribute depending on how it was created.
        """
        s = Store()
        q = s.query(C, limit=3)
        self.assertEqual(q.limit, 3)


    def test_itemQueryClone(self):
        """
        L{ItemQuery.cloneQuery} should return a new L{ItemQuery} which is
        equivalent, but not identical, to the original query.
        """
        s = Store()
        q1 = s.query(C, limit=3)
        q2 = q1.cloneQuery()
        self.assertEqual(q1.store, q2.store)
        self.assertEqual(q2.limit, q1.limit)
        self.assertNotIdentical(q1, q2)


    def test_itemQueryLimitSetToNone(self):
        """
        L{ItemQuery.cloneQuery} should set the limit attribute back to None when
        None is passed.  (e.g. it should not use None as the default value.
        HAX!)
        """
        s = Store()
        q1 = s.query(C, limit=3)
        q2 = q1.cloneQuery(limit=None)
        self.assertEqual(q1.store, q2.store)
        self.assertEqual(q2.limit, None)


    def test_itemQueryCloneLimit(self):
        """
        L{ItemQuery} implements L{IQuery} and should provide a 'cloneQuery'
        method which can accept a 'limit' parameter to change its limit.
        """
        s = Store()
        q1 = s.query(C, limit=3)
        q2 = q1.cloneQuery(5)
        self.assertEqual(q1.store, q2.store)
        self.assertEqual(q2.limit, 5)


    def test_itemQueryStoreAttribute(self):
        """
        L{ItemQuery} implements L{IQuery} and should provide a 'store' attribute
        that points to the store it will yield items from.
        """
        s = Store()
        q = s.query(C)
        self.assertEqual(q.store, s)


    def test_itemQueryDistinctLimitAttribute(self):
        """
        L{_DistinctQuery} implements L{IQuery} and should provide a 'limit'
        attribute depending on how it was created.
        """
        s = Store()
        q = s.query(C, limit=3).distinct()
        self.assertEqual(q.limit, 3)


    def test_itemQueryDistinctClone(self):
        """
        L{_DistinctQuery.cloneQuery} should return a new L{ItemQuery} which is
        equivalent, but not identical, to the original query.
        """
        s = Store()
        q1 = s.query(C, limit=3).distinct()
        q2 = q1.cloneQuery()
        self.assertEqual(q1.store, q2.store)
        self.assertEqual(q2.limit, q1.limit)


    def test_itemQueryDistinctCloneLimit(self):
        """
        L{_DistinctQuery} implements L{IQuery} and should provide a 'cloneQuery'
        method which can accept a 'limit' parameter to change its limit.
        """
        s = Store()
        q1 = s.query(C, limit=3).distinct()
        q2 = q1.cloneQuery(5)
        self.assertEqual(q1.store, q2.store)
        self.assertEqual(q2.limit, 5)


    def test_itemQueryDistinctStoreAttribute(self):
        """
        L{_DistinctQuery} implements L{IQuery} and should provide a 'store' attribute
        that points to the store it will yield items from.
        """
        s = Store()
        q = s.query(C).distinct()
        self.assertEqual(q.store, s)



    def testAttributeQueryMinMax(self):
        s = Store()
        def entesten():
            E(store=s, amount=-4)
            E(store=s, amount=10)
            E(store=s, amount=99)
            E(store=s, amount=456)

            self.assertEquals(s.query(E).getColumn("amount").min(), -4)
            self.assertEquals(s.query(E).getColumn("amount").max(), 456)

            self.assertRaises(ValueError, s.query(D).getColumn("id").max)
            self.assertRaises(ValueError, s.query(D).getColumn("id").min)

            self.assertEquals(s.query(D).getColumn("id").min(default=41), 41)
            self.assertEquals(s.query(D).getColumn("id").max(default=42), 42)


        s.transact(entesten)
        s.close()

    def test_attributeQueryLimitAttribute(self):
        """
        L{AttributeQuery} implements LI{IQuery} and should provide a 'limit'
        attribute depending on how it was created.
        """
        s = Store()
        q = s.query(C, limit=3).getColumn('name')
        self.assertEqual(q.limit, 3)


    def test_attributeQueryClone(self):
        """
        L{AttributeQuery.cloneQuery} should return a new L{AttributeQuery}
        which is equivalent, but not identical, to the original query.
        """
        s = Store()
        q1 = s.query(C, limit=3).getColumn('name')
        q2 = q1.cloneQuery()
        self.assertEqual(q1.store, q2.store)
        self.assertEqual(q2.limit, q1.limit)


    def test_attributeQueryCloneLimit(self):
        """
        L{AttributeQuery} implements L{IQuery} and should provide a 'cloneQuery'
        method which can accept a 'limit' parameter to change its limit.
        """
        s = Store()
        q1 = s.query(C, limit=3).getColumn('name')
        q2 = q1.cloneQuery(5)
        self.assertEqual(q1.store, q2.store)
        self.assertEqual(q2.limit, 5)
        self.assertEqual(q1.attribute, q2.attribute)
        self.assertEqual(q1.raw, q2.raw)


    def test_attributeQueryStoreAttribute(self):
        """
        L{AttributeQuery} implements L{IQuery} and should provide a 'store' attribute
        that points to the store it will yield attributes from.
        """
        s = Store()
        q = s.query(C).getColumn('name')
        self.assertEqual(q.store, s)



class MultipleQuery(TestCase):
    """
    Test cases for queries that yield multiple item types.
    """

    def test_basicJoin(self):
        """
        Verify that querying for multiple Item classes gives
        us the right number and type of Items, in a plausible order.
        """
        s = Store()
        def entesten():
            for i in range(3):
                c = C(store=s, name=u"C.%s" % i)
                B(store=s, name=u"B.%s" % (2-i), cref=c)

            query = s.query( (B, C),
                             B.cref == C.storeID,
                             sort=C.name.ascending)

            self.assertEquals(query.count(), 3)

            result = iter(query).next()

            self.assertEquals(len(result), 2)

            b, c = result

            self.assertTrue(isinstance(b, B))
            self.assertTrue(isinstance(c, C))

            self.assertEquals(b.name, u"B.2")
            self.assertEquals(c.name, u"C.0")

        s.transact(entesten)
        s.close()

    def test_count(self):
        """
        Verify that count() gives the right result in the
        presence of offset and limit.
        """
        s = Store()
        def entesten():
            for i in range(3):
                c = C(store=s, name=u"C.%s" % i)
                B(store=s, name=u"B.%s" % (2-i), cref=c)
                B(store=s, name=u"B2.%s" % (2-i), cref=c)

            query = s.query( (B, C),
                             B.cref == C.storeID)

            totalCombinations = 6

            self.assertEquals(query.count(), totalCombinations)

            for offset in range(totalCombinations):
                for limit in range(totalCombinations + 1):
                    query = s.query( (B, C),
                                     B.cref == C.storeID,
                                     offset=offset, limit=limit )
                    expectedCount = min((totalCombinations-offset), limit)
                    actualCount = query.count()
                    self.assertEquals(actualCount, expectedCount,
                                      "Got %s results with offset %s, limit %s" % (
                                      actualCount, offset, limit))

        s.transact(entesten)
        s.close()

    def test_distinct(self):
        """
        Verify that distinct gives the right answers for a multiple
        item queries.
        """
        s = Store()
        def entesten():
            for i in range(3):
                c = C(store=s, name=u"C.%s" % i)
                b = B(store=s, name=u"B.%s" % i, cref=c)
                A(store=s, type=u"A.%s" % i, reftoc=b)
                A(store=s, type=u"A.%s" % i, reftoc=b)

            query = s.query( (B, C),
                             AND(B.cref == C.storeID,
                                 A.reftoc == B.storeID),
                             sort = C.name.ascending )

            self.assertEquals(query.count(), 6)

            distinct = query.distinct()

            self.assertEquals(distinct.count(), 3)

            for i, (b, c) in enumerate(query.distinct()):
                self.assertEquals(b.name, u"B.%s" % i)
                self.assertEquals(c.name, u"C.%s" % i)

        s.transact(entesten)
        s.close()

    def test_tree(self):
        """
        Verify that queries using the same Item class more than
        once behave correctly.
        """
        s = Store()
        def entesten():
            pops = B(store=s, name=u"Pops")
            dad = B(store=s, name=u"Dad", cref=pops)
            B(store=s, name=u"Bro", cref=dad)
            B(store=s, name=u"Sis", cref=dad)

            Gen1 = Placeholder(B)
            Gen2 = Placeholder(B)
            Gen3 = Placeholder(B)

            query = s.query( (Gen1, Gen2, Gen3),
                             AND(Gen3.cref == Gen2.storeID,
                                 Gen2.cref == Gen1.storeID),
                             sort=Gen3.name.ascending )

            self.assertEquals(query.count(), 2)

            self.assertEquals(tuple(b.name for b in iter(query).next()),
                              (u"Pops", u"Dad", u"Bro"))
        s.transact(entesten)
        s.close()

    def test_oneTuple(self):
        """
        Verify that tuples of length one don't do anything crazy.
        """
        s = Store()
        def entesten():
            for i in range(3):
                C(store=s, name=u"C.%s" % i)

            query = s.query( (C,),
                             sort=C.name.ascending)

            self.assertEquals(query.count(), 3)

            results = iter(query)

            for i in range(3):
                result = results.next()
                self.assertEquals(len(result), 1)
                c, = result
                self.assertTrue(isinstance(c, C), "%s is not a C" % c)
                self.assertEquals(c.name, u"C.%s" % i, i)

        s.transact(entesten)
        s.close()

    def test_emptyTuple(self):
        """
        Verify that empty tuples don't give SQL crashes.
        """
        s = Store()
        def entesten():
            self.assertRaises(ValueError, s.query, ())

        s.transact(entesten)
        s.close()


    def test_limitAttribute(self):
        """
        L{MultipleItemQuery} implements LI{IQuery} and should provide a 'limit'
        attribute depending on how it was created.
        """
        s = Store()
        q = s.query((B, C), limit=3)
        self.assertEqual(q.limit, 3)


    def test_clone(self):
        """
        L{MultipleItemQuery.cloneQuery} should return a new L{MultipleItemQuery}
        which is equivalent, but not identical, to the original query.
        """
        s = Store()
        q1 = s.query((B, C), limit=3)
        q2 = q1.cloneQuery()
        self.assertEqual(q1.store, q2.store)
        self.assertEqual(q2.limit, q1.limit)


    def test_cloneLimit(self):
        """
        L{MultipleItemQuery} implements L{IQuery} and should provide a 'cloneQuery'
        method which can accept a 'limit' parameter to change its limit.
        """
        s = Store()
        q1 = s.query((B, C), limit=3)
        q2 = q1.cloneQuery(5)
        self.assertEqual(q1.store, q2.store)
        self.assertEqual(q2.limit, 5)


    def test_storeAttribute(self):
        """
        L{MultipleItemQuery} implements L{IQuery} and should provide a 'store' attribute
        that points to the store it will yield attributes from.
        """
        s = Store()
        q = s.query((B, C))
        self.assertEqual(q.store, s)


    def test_limitAttributeDistinct(self):
        """
        L{_MultipleItemDistinctQuery} implements LI{IQuery} and should provide a
        'limit' attribute depending on how it was created.
        """
        s = Store()
        q = s.query((B, C), limit=3)
        self.assertEqual(q.limit, 3)


    def test_cloneDistinct(self):
        """
        L{_MultipleItemDistinctQuery.cloneQuery} should return a new
        L{MultipleItemQuery} which is equivalent, but not identical, to the
        original query.
        """
        s = Store()
        q1 = s.query((B, C), limit=3).distinct()
        q2 = q1.cloneQuery()
        self.assertEqual(q1.store, q2.store)
        self.assertEqual(q2.limit, q1.limit)


    def test_cloneLimitDistinct(self):
        """
        L{_MultipleItemDistinctQuery} implements L{IQuery} and should provide a
        'cloneQuery' method which can accept a 'limit' parameter to change its
        limit.
        """
        s = Store()
        q1 = s.query((B, C), limit=3).distinct()
        q2 = q1.cloneQuery(5)
        self.assertEqual(q1.store, q2.store)
        self.assertEqual(q2.limit, 5)


    def test_storeAttributeDistinct(self):
        """
        L{_MultipleItemDistinctQuery} implements L{IQuery} and should provide a
        'store' attribute that points to the store it will yield attributes
        from.
        """
        s = Store()
        q = s.query((B, C)).distinct()
        self.assertEqual(q.store, s)



class QueryingTestCase(TestCase):
    def setUp(self):
        s = self.store = Store()
        def _createStuff():
            self.d1 = D(store=s, one='d1.one', two='d1.two', three='d1.three', four=u'd1.four', id='1')
            self.d2 = D(store=s, one='d2.one', two='d2.two', three='d2.three', four=u'd2.four', id='2')
            self.d3 = D(store=s, one='d3.one', two='d3.two', three='d3.three', four=u'd3.four', id='3')
        s.transact(_createStuff)

    def tearDown(self):
        self.store.close()

    def query(self, *a, **kw):
        return list(self.store.query(*a, **kw))

    def assertQuery(self, query, expected, args=None):
        """
        Perform byte-for-byte comparisons against generated SQL.  It would be
        slightly nicer if we have a SQL parser which emited an AST we could
        test against instead, but in the absence of that, we'll do the more
        difficult thing and keep the tests in sync with the SQL generator.
        If, someday, we have multiple backends which have different SQL
        generation requirements, we'll probably need to split all these tests
        up.

        While it is true that we don't actually directly care about what SQL
        gets generated, we do want to test the SQL generation as a /unit/,
        rather than indirectly testing it by making assertions about the
        result set it generates.  This for all the usual reasons one writes
        unit tests (ease of debugging, refactoring, maintenance).  Other
        tests cover the actual query behavior this SQL results in, and
        ideally some day we will have some tests which interact with the
        actual underlying rdbm to test basic assumptions we are making about
        the behavior of particular snippets of SQL.

        To sum up, changes to the SQL generation code may rightly require
        changes to tests which use assertQuery.  If the SQL we want to generate
        changes, do not be afraid to update the tests.
        """
        if args is None:
            args = []

        sql = query.getQuery(self.store)
        self.assertEquals(
            sql,
            expected,
            "\n%r != %r\n(if SQL generation code has changed, maybe this test "
            "should be updated)\n" % (sql, expected))
        self.assertEquals([str(a) for a in query.getArgs(self.store)], args)



class TableOrder(TestCase):
    """
    Tests for the order of tables when joins are being performed.
    """
    def test_baseQuery(self):
        """
        Test that the simplest query possible, one with only a FROM clause,
        specifies only the table for the type being queried.
        """
        store = Store()
        query = store.query(A)
        self.assertEqual(query.fromClauseParts, [store.getTableName(A)])


    def test_singleValueComparison(self):
        """
        Test that a query with a simple comparison against a value specifies
        only the table for the type being queried.
        """
        store = Store()
        query = store.query(A, A.type == u'value')
        self.assertEqual(query.fromClauseParts, [store.getTableName(A)])


    def test_twoColumnComparison(self):
        """
        Test that a query which compares one column against another from the
        type being queried specifies only the table for the type being queried
        for the FROM clause.
        """
        store = Store()
        query = store.query(A, A.type == A.reftoc)
        self.assertEqual(query.fromClauseParts, [store.getTableName(A)])


    def test_baseSort(self):
        """
        Test that a query which includes an ORDER BY clause including a column
        from the type being queried properly includes only the table for the
        type being queried in the FROM clause.
        """
        store = Store()
        query = store.query(A, sort=A.type.ascending)
        self.assertEqual(query.fromClauseParts, [store.getTableName(A)])


    def test_comparisonWithSort(self):
        """
        Test that a query with both a WHERE clause and an ORDER BY clause from
        the table being queried properly has only the table for the type being
        queried included in the FROM clause.
        """
        store = Store()
        query = store.query(A, A.type == u'value', sort=A.type.ascending)
        self.assertEqual(query.fromClauseParts, [store.getTableName(A)])


    def test_invalidComparison(self):
        """
        Test that a query with a WHERE clause which does not reference the type
        being queried for is rejected.
        """
        store = Store()
        self.assertRaises(
            ValueError,
            store.query, A, B.name == u'value')


    def test_invalidSort(self):
        """
        Test that sorting by a column from a table other than the one being
        queried without joining on that table is rejected.
        """
        store = Store()
        self.assertRaises(
            ValueError,
            store.query, A, sort=B.name.ascending)


    def test_singleJoin(self):
        """
        Test that joining on another table by performing a comparison between
        two types properly includes both table names, in the right order (the
        order of tables in the comparison, left to right), in the FROM CLAUSE.
        """
        store = Store()
        query = store.query(A, A.type == B.name)
        self.assertEqual(
            query.fromClauseParts,
            [store.getTableName(A), store.getTableName(B)])


    def test_singleJoinReversed(self):
        """
        Test that joining on another table by performing a comparison between
        two types properly includes both table names, in the right order (the
        order of tables in the comparison, left to right), in the FROM CLAUSE.
        """
        store = Store()
        query = store.query(A, B.name == A.type)
        self.assertEqual(
            query.fromClauseParts,
            [store.getTableName(B), store.getTableName(A)])


    def test_explicitTableOrder(self):
        """
        Test that the order of tables in a join can be explicitly specified by
        using L{TableOrderComparisonWrapper}.
        """
        store = Store()
        query = store.query(
            A, TableOrderComparisonWrapper(
                [E, D, C, B, A],
                AND(A.type == B.name,
                    B.name == C.name,
                    C.name == D.four,
                    D.four == E.name)))
        self.assertEqual(
            query.fromClauseParts,
            [store.getTableName(E), store.getTableName(D),
             store.getTableName(C), store.getTableName(B),
             store.getTableName(A)])


class AndOrQueries(QueryingTestCase):
    def testNoConditions(self):
        self.assertRaises(ValueError, AND)
        self.assertRaises(ValueError, OR)


    def testOneCondition(self):
        """
        Test that an L{AND} or an L{OR} with a single argument collapses to
        just that argument.
        """
        self.assertQuery(
            AND(A.type == u'Narf!'),
            '((%s = ?))' % (A.type.getColumnName(self.store),),
            ['Narf!'])
        self.assertQuery(
            OR(A.type == u'Narf!'),
            '((%s = ?))' % (A.type.getColumnName(self.store),),
            ['Narf!'])
        self.assertEquals(self.query(D, AND(D.one == 'd1.one')), [self.d1])
        self.assertEquals(self.query(D, OR(D.one == 'd1.one')), [self.d1])


    def testMultipleAndConditions(self):
        condition = AND(A.type == u'Narf!',
                        A.type == u'Poiuyt!',
                        A.type == u'Try to take over the world')

        expectedSQL = '((%s = ?) AND (%s = ?) AND (%s = ?))'
        expectedSQL %= (A.type.getColumnName(self.store),) * 3

        self.assertQuery(
            condition,
            expectedSQL,
            ['Narf!', 'Poiuyt!', 'Try to take over the world'])
        self.assertEquals(
            self.query(D, AND(D.one == 'd1.one',
                              D.two == 'd1.two',
                              D.three == 'd1.three')),
            [self.d1])

    def testMultipleOrConditions(self):
        condition = OR(A.type == u'Narf!',
                       A.type == u'Poiuyt!',
                       A.type == u'Try to take over the world')
        expectedSQL = '((%s = ?) OR (%s = ?) OR (%s = ?))'
        expectedSQL %= (A.type.getColumnName(self.store),) * 3
        self.assertQuery(
            condition,
            expectedSQL,
            ['Narf!', 'Poiuyt!', 'Try to take over the world'])
        q = self.query(D, OR(D.one == 'd1.one',
                             D.one == 'd2.one',
                             D.one == 'd3.one'))
        e = [self.d1, self.d2, self.d3]
        self.assertEquals(sorted(q), sorted(e))


class SetMembershipQuery(QueryingTestCase):

    def test_oneOfValueQueryGeneration(self):
        """
        Test that comparing an attribute for containment against a value set
        generates the appropriate SQL.
        """
        values = [u'a', u'b', u'c']
        comparison = C.name.oneOf(values)
        self.failUnless(IComparison.providedBy(comparison))
        self.assertEquals(
            comparison.getQuery(self.store),
            '%s IN (?, ?, ?)' % (
                C.name.getColumnName(self.store),))
        self.assertEquals(
            comparison.getArgs(self.store),
            values)


    def test_oneOfColumnQueryGeneration(self):
        """
        Test that comparing an attribute for containment against an L{IColumn}
        generates the appropriate SQL.
        """
        values = A.type
        comparison = C.name.oneOf(values)
        self.failUnless(IComparison.providedBy(comparison))
        self.assertEquals(
            comparison.getQuery(self.store),
            '%s IN (%s)' % (
                C.name.getColumnName(self.store),
                A.type.getColumnName(self.store)))
        self.assertEquals(
            comparison.getArgs(self.store),
            [])


    def test_oneOfColumnQueryQueryGeneration(self):
        """
        Test that comparing an attribute for containment against another query
        generates a sub-select.
        """
        subselect = self.store.query(A).getColumn('type')
        comparison = C.name.oneOf(subselect)
        self.failUnless(IComparison.providedBy(comparison))
        self.assertEquals(
            comparison.getQuery(self.store),
            '%s IN (SELECT %s FROM %s)' % (
                C.name.getColumnName(self.store),
                A.type.getColumnName(self.store),
                A.getTableName(self.store)))
        self.assertEquals(
            comparison.getArgs(self.store),
            [])


    def test_oneOfColumnQueryQueryGenerationWithArguments(self):
        """
        Like test_oneOfColumnQueryQueryGeneration, but pass some values to the
        subselect and make sure they come out of the C{getArgs} method
        properly.
        """
        value = '10'
        subselect = self.store.query(
            D,
            AND(D.id == value,
                D.four == C.name)).getColumn('one')

        comparison = C.name.oneOf(subselect)
        self.failUnless(IComparison.providedBy(comparison))
        self.assertEquals(
            comparison.getQuery(self.store),
            '%s IN (SELECT %s FROM %s, %s WHERE ((%s = ?) AND (%s = %s)))' % (
                C.name.getColumnName(self.store),
                D.one.getColumnName(self.store),
                D.getTableName(self.store),
                C.getTableName(self.store),
                D.id.getColumnName(self.store),
                D.four.getColumnName(self.store),
                C.name.getColumnName(self.store)))
        self.assertEquals(
            map(str, comparison.getArgs(self.store)),
            [value])


    def testOneOfWithList(self):
        cx = C(store=self.store, name=u'x')
        C(store=self.store, name=u'y')
        cz = C(store=self.store, name=u'z')

        query = self.store.query(
            C, C.name.oneOf([u'x', u'z', u'a']), sort=C.name.ascending)

        self.assertEquals(list(query), [cx, cz])


    def testOneOfWithSet(self):
        s = Store()

        cx = C(store=s, name=u'x')
        C(store=s, name=u'y')
        cz = C(store=s, name=u'z')

        self.assertEquals(list(s.query(C, C.name.oneOf(set([u'x', u'z', u'a'])), sort=C.name.ascending)),
                          [cx, cz])


class WildcardQueries(QueryingTestCase):
    def testNoConditions(self):
        self.assertRaises(TypeError, D.one.like)
        self.assertRaises(TypeError, D.one.notLike)


    def test_likeValueComparisonInvolvedTables(self):
        """
        Test that only the table to which a column belongs is included in the
        involved tables set when that column is compared to a literal value
        using like.
        """
        comparison = D.one.like('foo')
        self.assertEqual(comparison.getInvolvedTables(), [D])


    def test_likeColumnComparisonInvolvedTables(self):
        """
        Test that both the table to which a column belongs and the table to
        which a target column belongs are included in the involved tables set
        when those columns are compared using like.
        """
        comparison = D.one.like(A.type)
        self.assertEqual(comparison.getInvolvedTables(), [D, A])


    def testOneString(self):
        self.assertQuery(
            D.one.like('foobar%'),
            '(%s LIKE (?))' % (D.one.getColumnName(self.store),),
            ['foobar%'])
        self.assertQuery(
            D.one.notLike('foobar%'),
            '(%s NOT LIKE (?))' % (D.one.getColumnName(self.store),),
            ['foobar%'])
        self.assertEquals(self.query(D, D.four.like(u'd1.four')), [self.d1])
        self.assertEquals(self.query(D, D.four.notLike(u'd%.four')), [])

    def testOneColumn(self):
        self.assertQuery(
            D.one.like(D.two),
            '(%s LIKE (%s))' % (D.one.getColumnName(self.store),
                                D.two.getColumnName(self.store)))
        self.assertEquals(self.query(D, D.one.like(D.two)), [])

    def testOneColumnAndStrings(self):
        self.assertQuery(
            D.four.like(u'%', D.id, u'%four'),
            '(%s LIKE (? || %s || ?))' % (D.four.getColumnName(self.store),
                                          D.id.getColumnName(self.store)),
            [u'%', u'%four'])
        q = self.query(D, D.four.like(u'%', D.id, u'%four'))
        e = [self.d1, self.d2, self.d3]
        self.assertEquals(sorted(q), sorted(e))

    def testMultipleColumns(self):
        self.assertQuery(
            D.one.like(D.two, '%', D.three),
            '(%s LIKE (%s || ? || %s))' % (D.one.getColumnName(self.store),
                                           D.two.getColumnName(self.store),
                                           D.three.getColumnName(self.store)),
            ['%'])
        self.assertEquals(
            self.query(D, D.one.like(D.two, '%', D.three)), [])


    def testStartsEndsWith(self):
        self.assertQuery(
            D.four.startswith(u'foo'),
            '(%s LIKE (?))' % (D.four.getColumnName(self.store),),
            ['foo%'])
        self.assertQuery(
            D.four.endswith(u'foo'),
            '(%s LIKE (?))' % (D.four.getColumnName(self.store),),
            ['%foo'])
        self.assertEquals(
            self.query(D, D.four.startswith(u'd1')), [self.d1])
        self.assertEquals(
            self.query(D, D.four.endswith(u'3.four')), [self.d3])


    def testStartsEndsWithText(self):
        self.assertEquals(
            self.query(D, D.four.startswith(u'd1')),
            [self.d1])
        self.assertEquals(
            self.query(D, D.four.endswith(u'2.four')),
            [self.d2])



class UniqueTest(TestCase):

    def setUp(self):
        s = self.s = Store()
        self.c = C(store=s, name=u'unique')
        self.dupc1 = C(store=s, name=u'non-unique')
        self.dupc2 = C(store=s, name=u'non-unique')

    def testUniqueFound(self):
        self.assertEquals(self.s.findUnique(C, C.name == u'unique'), self.c)

    def testUniqueNotFoundError(self):
        self.assertRaises(errors.ItemNotFound, self.s.findUnique,
                          C, C.name == u'non-existent')

    def testUniqueNotFoundDefault(self):
        bing = object()
        self.assertEquals(bing, self.s.findUnique(
                C, C.name == u'non-existent',
                default=bing))

    def testUniqueDuplicate(self):
        self.assertRaises(errors.DuplicateUniqueItem,
                          self.s.findUnique, C, C.name == u'non-unique')



class PlaceholderTestItem(Item):
    """
    Type used by the placeholder support test cases.
    """
    attr = integer()
    other = integer()
    characters = text()


COMPARISON_OPS = [
    operator.lt, operator.le, operator.eq,
    operator.ne, operator.ge, operator.gt]

class PlaceholderTestCase(TestCase):
    """
    Tests for placeholder table name support.
    """
    def test_placeholderType(self):
        """
        Test that the C{type} attribute of a Placeholder column is the
        Placeholder from which it came.
        """
        p = Placeholder(PlaceholderTestItem)
        a = p.attr
        self.assertIdentical(a.type, p)


    def test_placeholderTableName(self):
        """
        Test that the table name of a Placeholder is the same as the table name
        of the underlying Item class.
        """
        s = Store()
        p = Placeholder(PlaceholderTestItem)
        self.assertEquals(p.getTableName(s), PlaceholderTestItem.getTableName(s))


    def test_placeholderColumnInterface(self):
        """
        Test that a column from a placeholder provides L{IColumn}.
        """
        p = Placeholder(PlaceholderTestItem)
        a = p.attr
        self.failUnless(IColumn.providedBy(a))


    def test_placeholderAttributeValueComparison(self):
        """
        Test that getting an attribute from a Placeholder which exists on the
        underlying Item class and comparing it to a value returns an
        L{IComparison} provider.
        """
        value = 0
        p = Placeholder(PlaceholderTestItem)
        for op in COMPARISON_OPS:
            self.failUnless(IComparison.providedBy(op(p.attr, value)))
            self.failUnless(IComparison.providedBy(op(value, p.attr)))


    def test_placeholderAttributeColumnComparison(self):
        """
        Test that getting an attribute from a Placeholder which exists on the
        underlying Item class and comparing it to another column returns an
        L{IComparison} provider.
        """
        p = Placeholder(PlaceholderTestItem)
        for op in COMPARISON_OPS:
            self.failUnless(IComparison.providedBy(op(p.attr, PlaceholderTestItem.attr)))
            self.failUnless(IComparison.providedBy(op(PlaceholderTestItem.attr, p.attr)))


    def _placeholderAttributeSimilarity(self, kind, sql, args):
        s = Store()
        value = u'text'

        p = Placeholder(PlaceholderTestItem)

        # Explicitly call this, since we aren't going through ItemQuery.
        p.getTableAlias(s, ())

        comparison = getattr(p.characters, kind)(value)
        self.failUnless(IComparison.providedBy(comparison))
        self.assertEquals(comparison.getQuery(s),
                          sql % (p.characters.getColumnName(s),))
        self.assertEquals(
            comparison.getArgs(s),
            [args % (value,)])


    def test_placeholderAttributeSimilarity(self):
        """
        Test that placeholder attributes can be used with the SQL LIKE
        operator.
        """
        return self._placeholderAttributeSimilarity('like', '(%s LIKE (?))', '%s')


    def test_placeholderAttributeDisimilarity(self):
        """
        Test that placeholder attributes can be used with the SQL NOT LIKE
        operator.
        """
        return self._placeholderAttributeSimilarity('notLike', '(%s NOT LIKE (?))', '%s')


    def test_placeholderAttributeStartsWith(self):
        """
        Test that placeholder attributes work with the .startswith() method.
        """
        return self._placeholderAttributeSimilarity('startswith', '(%s LIKE (?))', '%s%%')


    def test_placeholderAttributeEndsWith(self):
        """
        Test that placeholder attributes work with the .endswith() method.
        """
        return self._placeholderAttributeSimilarity('endswith', '(%s LIKE (?))', '%%%s')


    def test_placeholderLikeTarget(self):
        """
        Test that a placeholder can be used as the right-hand argument to a SQL
        LIKE expression.
        """
        s = Store()
        p = Placeholder(PlaceholderTestItem)

        # Call this since we're not using ItemQuery
        p.getTableAlias(s, ())

        comparison = PlaceholderTestItem.attr.like(p.attr)
        self.failUnless(IComparison.providedBy(comparison))
        self.assertEquals(
            comparison.getQuery(s),
            '(%s LIKE (placeholder_0.[attr]))' % (
                PlaceholderTestItem.attr.getColumnName(s),))
        self.assertEquals(
            comparison.getArgs(s),
            [])


    def test_placeholderContainment(self):
        """
        Test that placeholder attributes can be used with the SQL IN and NOT IN
        operators.
        """
        s = Store()
        value = [1, 2, 3]
        p = Placeholder(PlaceholderTestItem)

        # Call this since we're not using ItemQuery
        p.getTableAlias(s, ())

        comparison = p.attr.oneOf(value)
        self.failUnless(IComparison.providedBy(comparison))
        self.assertEquals(
            comparison.getQuery(s),
            '%s IN (?, ?, ?)' % (p.attr.getColumnName(s),))
        self.assertEquals(
            comparison.getArgs(s),
            value)


    def test_placeholderAntiContainment(self):
        """
        Test that placeholder attributes can be used with the SQL NOT IN
        operator.
        """
        s = Store()
        value = [1, 2, 3]
        p = Placeholder(PlaceholderTestItem)

        # Call this since we're not using ItemQuery
        p.getTableAlias(s, ())

        comparison = p.attr.notOneOf(value)
        self.failUnless(IComparison.providedBy(comparison))
        self.assertEquals(
            comparison.getQuery(s),
            '%s NOT IN (?, ?, ?)' % (p.attr.getColumnName(s),))
        self.assertEquals(
            comparison.getArgs(s),
            value)


    def test_placeholderContainmentTarget(self):
        """
        Test that a placeholder attribute can be used as the right-hand
        argument to the SQL IN operator.
        """
        s = Store()
        p = Placeholder(PlaceholderTestItem)

        # Call this since we're not using ItemQuery
        p.getTableAlias(s, ())

        comparison = PlaceholderTestItem.attr.oneOf(p.attr)
        self.failUnless(IComparison.providedBy(comparison))
        self.assertEquals(
            comparison.getQuery(s),
            '%s IN (%s)' % (PlaceholderTestItem.attr.getColumnName(s),
                            p.attr.getColumnName(s)))
        self.assertEquals(
            comparison.getArgs(s),
            [])


    def test_placeholderAntiContainmentTarget(self):
        """
        Test that a placeholder attribute can be used as the right-hand
        argument to the SQL NOT IN operator.
        """
        s = Store()
        p = Placeholder(PlaceholderTestItem)

        # Call this since we're not using ItemQuery
        p.getTableAlias(s, ())

        comparison = PlaceholderTestItem.attr.notOneOf(p.attr)
        self.failUnless(IComparison.providedBy(comparison))
        self.assertEquals(
            comparison.getQuery(s),
            '%s NOT IN (%s)' % (PlaceholderTestItem.attr.getColumnName(s),
                            p.attr.getColumnName(s)))
        self.assertEquals(
            comparison.getArgs(s),
            [])


    def test_placeholderStoreID(self):
        """
        Test that the C{storeID} attribute of a Placeholder can be retrieved
        just like any other attribute.
        """
        value = 0
        p = Placeholder(PlaceholderTestItem)
        self.failUnless(IComparison.providedBy(p.storeID > value))


    def test_placeholderAttributeError(self):
        """
        Test that trying to get an attribute from a Placeholder which is not an
        L{IComparison} on the underlying Item class raises an AttributeError.
        """
        p = Placeholder(PlaceholderTestItem)
        self.assertRaises(AttributeError, getattr, p, 'nonexistentAttribute')


    def test_placeholderComparisonTables(self):
        """
        Test that the result of L{IComparison.getInvolvedTables} on an
        attribute retrieved from a Placeholder returns a special placeholder
        item.
        """
        s = Store()
        p = Placeholder(PlaceholderTestItem)
        value = 0
        involvedTables = (p.attr > value).getInvolvedTables()
        self.assertEquals(len(involvedTables), 1)
        theTable = iter(involvedTables).next()
        self.assertEquals(theTable.getTableName(s),
                          PlaceholderTestItem.getTableName(s))
        self.assertEquals(theTable.getTableAlias(s, ()),
                          'placeholder_0')


    def test_placeholderComparisonQuery(self):
        """
        Test that the result of L{IComparison.getQuery} on an attribute
        retrieved from a Placeholder returns SQL which correctly uses an alias
        of the wrapped table.
        """
        s = Store()
        p = Placeholder(PlaceholderTestItem)

        # Explicitly call this here, since we're not going through ItemQuery or
        # another more reasonable codepath, which would have called it for us.
        p.getTableAlias(s, ())

        value = 0
        comparison = (p.attr > value)
        self.assertEquals(
            comparison.getQuery(s),
            '(placeholder_0.[attr] > ?)')
        self.assertEquals(
            comparison.getArgs(s),
            [value])


    def test_placeholderComparisonArgs(self):
        """
        Test that the result of L{IComparison.getArgs} on an attribute
        retrieved from a Placeholder returns the right values for the
        comparison.
        """
        s = Store()
        p = Placeholder(PlaceholderTestItem)
        value = 0
        args = (p.attr > value).getArgs(s)
        self.assertEquals(args, [0])


    def test_placeholderQuery(self):
        """
        Test that an ItemQuery can be created with Placeholder instances and
        the SQL it emits as a result correctly assigns and uses table aliases.
        """
        s = Store()
        p = Placeholder(PlaceholderTestItem)
        sql, args = ItemQuery(s, p)._sqlAndArgs('SELECT', '*')
        self.assertEquals(
            sql,
            'SELECT * FROM %s AS placeholder_0' % (
                PlaceholderTestItem.getTableName(s),))


    def test_placeholderMultiQuery(self):
        """
        Test that a MultipleItemQuery can be created with Placeholder instances
        and the SQL it emits as a result correctly assigns and uses table
        aliases.
        """
        s = Store()
        p1 = Placeholder(PlaceholderTestItem)
        p2 = Placeholder(PlaceholderTestItem)
        sql, args = MultipleItemQuery(s, (p1, p2))._sqlAndArgs('SELECT', '*')
        tableName = PlaceholderTestItem.getTableName(s)
        self.assertEqual(
            sql,
            'SELECT * FROM %s AS placeholder_0, %s AS placeholder_1' % (
                tableName, tableName))


    def test_placeholderComparison(self):
        """
        Test that a comparison which contains a Placeholder also results in
        properly generated SQL.
        """
        s = Store()
        p = Placeholder(PlaceholderTestItem)
        query = ItemQuery(
            s,
            PlaceholderTestItem,
            PlaceholderTestItem.attr == p.attr)
        sql, args = query._sqlAndArgs('SELECT', '*')
        self.assertEquals(
            sql,
            'SELECT * '
            'FROM %s, %s AS placeholder_0 '
            'WHERE (%s.[attr] = placeholder_0.[attr])' % (
                PlaceholderTestItem.getTableName(s),
                PlaceholderTestItem.getTableName(s),
                PlaceholderTestItem.getTableName(s)))
        self.assertEquals(args, [])


    def test_placeholderOrdering(self):
        """
        Placeholders should be ordered based on the order in which they were
        instantiated.
        """
        p1 = Placeholder(PlaceholderTestItem)
        p2 = Placeholder(PlaceholderTestItem)
        self.failUnless(p1 < p2)
        self.failUnless(p2 > p1)
        self.failIf(p1 >= p2)
        self.failIf(p2 <= p1)
        self.failIf(p1 == p2)
        self.failIf(p2 == p1)
        self.failUnless(p1 != p2)
        self.failUnless(p2 != p1)


    def test_placeholderObjectSorting(self):
        """
        Placeholders should sort based on the order in which they were
        instantiated.
        """
        placeholders = [Placeholder(PlaceholderTestItem) for n in xrange(10)]
        shuffledPlaceholders = list(placeholders)
        random.shuffle(shuffledPlaceholders)
        shuffledPlaceholders.sort()
        self.assertEquals(placeholders, shuffledPlaceholders)


    def test_placeholderAliasAssignment(self):
        """
        Test that each placeholder selects a unique alias for itself.
        """
        s = Store()
        p1 = Placeholder(PlaceholderTestItem)
        p2 = Placeholder(PlaceholderTestItem)

        aliases = []
        self.assertEquals(p1.getTableAlias(s, aliases), 'placeholder_0')
        self.assertEquals(p1.getTableAlias(s, aliases), 'placeholder_0')
        aliases.append('placeholder_')
        self.assertEquals(p1.getTableAlias(s, aliases), 'placeholder_0')
        self.assertEquals(p2.getTableAlias(s, aliases), 'placeholder_1')


    def test_multiplePlaceholderComparisons(self):
        """
        Test that using multiple different placeholders in a comparison at once
        properly gives each a unique name.
        """
        s = Store()
        p1 = Placeholder(PlaceholderTestItem)
        p2 = Placeholder(PlaceholderTestItem)

        query = ItemQuery(
            s,
            PlaceholderTestItem,
            AND(PlaceholderTestItem.attr == p1.attr,
                PlaceholderTestItem.other == p1.other,
                PlaceholderTestItem.attr == p2.attr,
                PlaceholderTestItem.characters == p2.characters))
        sql, args = query._sqlAndArgs('SELECT', '*')
        self.assertEquals(
            sql,
            'SELECT * '
            'FROM %s, %s AS placeholder_0, %s AS placeholder_1 '
            'WHERE ((%s = placeholder_0.[attr]) AND '
                   '(%s = placeholder_0.[other]) AND '
                   '(%s = placeholder_1.[attr]) AND '
                   '(%s = placeholder_1.[characters]))' % (
                PlaceholderTestItem.getTableName(s),
                PlaceholderTestItem.getTableName(s),
                PlaceholderTestItem.getTableName(s),
                PlaceholderTestItem.attr.getColumnName(s),
                PlaceholderTestItem.other.getColumnName(s),
                PlaceholderTestItem.attr.getColumnName(s),
                PlaceholderTestItem.characters.getColumnName(s)))
        self.assertEquals(args, [])


    def test_sortByPlaceholderAttribute(self):
        """
        Test that a placeholder attribute can be used as a sort key.
        """
        s = Store()
        p = Placeholder(PlaceholderTestItem)

        query = ItemQuery(
            s,
            p,
            sort=p.attr.ascending)
        sql, args = query._sqlAndArgs('SELECT', '*')

        expectedSQL = ('SELECT * '
                       'FROM %s AS placeholder_0 '
                       'ORDER BY placeholder_0.[attr] ASC')
        expectedSQL %= (p.getTableName(s),)

        self.assertEquals(sql, expectedSQL)
        self.assertEquals(args, [])


    def test_placeholderColumnNamesInQueryTarget(self):
        """
        Test that placeholders are used correctly in the 'result'
        portion of an SQL query.
        """
        s = Store()
        p = Placeholder(PlaceholderTestItem)

        query = ItemQuery(s, p)

        expectedSQL = "placeholder_0.oid, placeholder_0.[attr], placeholder_0.[characters], placeholder_0.[other]"

        self.assertEquals(query._queryTarget, expectedSQL)



class BytesDeprecatedLikeTests(TestCase):
    """
    Deprecated tests for LIKE queries on L{axiom.attributes.bytes}.
    """
    def test_startsWith(self):
        self.assertWarns(
            DeprecationWarning,
            'axiom.attributes.startswith was deprecated in Axiom 0.7.5',
            __file__,
            lambda: D.one.startswith('string'))


    def test_endsWith(self):
        self.assertWarns(
            DeprecationWarning,
            'axiom.attributes.endswith was deprecated in Axiom 0.7.5',
            __file__,
            lambda: D.one.endswith('string'))


    def test_like(self):
        self.assertWarns(
            DeprecationWarning,
            'axiom.attributes.like was deprecated in Axiom 0.7.5',
            __file__,
            lambda: D.one.like('string'))


    def test_notLike(self):
        self.assertWarns(
            DeprecationWarning,
            'axiom.attributes.notLike was deprecated in Axiom 0.7.5',
            __file__,
            lambda: D.one.notLike('string'))



from weakref import ref
from traceback import print_exc

from twisted.python import log

from axiom import iaxiom

class CacheFault(KeyError):
    """
    An item has fallen out of cache, but the weakref callback has not yet run.
    """



class CacheInconsistency(RuntimeError):
    """
    A key being cached is already present in the cache.
    """



def logErrorNoMatterWhat():
    try:
        log.msg("Exception in finalizer cannot be propagated")
        log.err()
    except:
        try:
            emergLog = file("WEAKREF_EMERGENCY_ERROR.log", 'a')
            print_exc(file=emergLog)
            emergLog.flush()
            emergLog.close()
        except:
            # Nothing can be done.  We can't get an emergency log file to write
            # to.  Don't bother.
            return



def createCacheRemoveCallback(cacheRef, key, finalizer):
    """
    Construct a callable to be used as a weakref callback for cache entries.

    The callable will invoke the provided finalizer, as well as removing the
    cache entry if the cache still exists and contains an entry for the given
    key.

    @type  cacheRef: L{weakref.ref} to L{FinalizingCache}
    @param cacheRef: A weakref to the cache in which the corresponding cache
        item was stored.

    @param key: The key for which this value is cached.

    @type  finalizer: callable taking 0 arguments
    @param finalizer: A user-provided callable that will be called when the
        weakref callback runs.
    """
    def remove(reference):
        # Weakref callbacks cannot raise exceptions or DOOM ensues
        try:
            finalizer()
        except:
            logErrorNoMatterWhat()
        try:
            cache = cacheRef()
            if cache is not None:
                if key in cache.data:
                    if cache.data[key] is reference:
                        del cache.data[key]
        except:
            logErrorNoMatterWhat()
    return remove



class FinalizingCache:
    """
    A cache that stores values by weakref.

    A finalizer is invoked when the weakref to a cached value is broken.

    @type data: L{dict}
    @ivar data: The cached values.
    """
    def __init__(self, _ref=ref):
        self.data = {}
        self._ref = _ref


    def cache(self, key, value):
        """
        Add an entry to the cache.

        A weakref to the value is stored, rather than a direct reference. The
        value must have a C{__finalizer__} method that returns a callable which
        will be invoked when the weakref is broken.

        @param key: The key identifying the cache entry.

        @param value: The value for the cache entry.
        """
        fin = value.__finalizer__()
        try:
            # It's okay if there's already a cache entry for this key as long
            # as the weakref has already been broken. See the comment in
            # get() for an explanation of why this might happen.
            if self.data[key]() is not None:
                raise CacheInconsistency(
                    "Duplicate cache key: %r %r %r" % (
                        key, value, self.data[key]))
        except KeyError:
            pass
        callback = createCacheRemoveCallback(self._ref(self), key, fin)
        self.data[key] = self._ref(value, callback)
        return value


    def uncache(self, key, value):
        """
        Remove a key from the cache.

        As a sanity check, if the specified key is present in the cache, it
        must have the given value.

        @param key: The key to remove.

        @param value: The expected value for the key.
        """
        try:
            assert self.get(key) is value
            del self.data[key]
        except KeyError:
            # If the entry has already been removed from the cache, this will
            # result in KeyError which we ignore. If the entry is still in the
            # cache, but the weakref has been broken, this will result in
            # CacheFault (a KeyError subclass) which we also ignore. See the
            # comment in get() for an explanation of why this might happen.
            pass


    def get(self, key):
        """
        Get an entry from the cache by key.

        @raise KeyError: if the given key is not present in the cache.

        @raise CacheFault: (a L{KeyError} subclass) if the given key is present
            in the cache, but the value it points to is gone.
        """
        o = self.data[key]()
        if o is None:
            # On CPython, the weakref callback will always(?) run before any
            # other code has a chance to observe that the weakref is broken;
            # and since the callback removes the item from the dict, this
            # branch of code should never run. However, on PyPy (and possibly
            # other Python implementations), the weakref callback does not run
            # immediately, thus we may be able to observe this intermediate
            # state. Should this occur, we remove the dict item ourselves,
            # and raise CacheFault (which is a KeyError subclass).
            del self.data[key]
            raise CacheFault(
                "FinalizingCache has %r but its value is no more." % (key,))
        log.msg(interface=iaxiom.IStatEvent, stat_cache_hits=1, key=key)
        return o




# -*- test-case-name: axiom.test.test_slotmachine -*-

hyper = super

_NOSLOT = object()

class Allowed(object):
    """
    An attribute that's allowed to be set.
    """
    def __init__(self, name, default=_NOSLOT):
        self.name = name
        self.default = default

    def __get__(self, oself, otype=None):
        if otype is not None and oself is None:
            return self
        if self.name in oself.__dict__:
            return oself.__dict__[self.name]
        if self.default is not _NOSLOT:
            return self.default
        raise AttributeError("%r object did not have attribute %r" %(oself.__class__.__name__, self.name))

    def __delete__(self, oself):
        if self.name not in oself.__dict__:
            # Returning rather than raising here because that's what
            # member_descriptor does, and Axiom relies upon that behavior.
            ## raise AttributeError('%r object has no attribute %r' %
            ##                      (oself.__class__.__name__, self.name))
            return
        del oself.__dict__[self.name]

    def __set__(self, oself, value):
        oself.__dict__[self.name] = value

class _SlotMetaMachine(type):
    def __new__(meta, name, bases, dictionary):
        dictionary['__name__'] = name
        slots = list(meta.determineSchema(dictionary))
        for slot in slots:
            for base in bases:
                defval = getattr(base, slot, _NOSLOT)
                if defval is not _NOSLOT:
                    break
            dictionary[slot] = Allowed(slot, defval)
        nt = type.__new__(meta, name, bases, dictionary)
        return nt

    def determineSchema(meta, dictionary):
        return dictionary.get("slots", [])

    determineSchema = classmethod(determineSchema)


class DescriptorWithDefault(object):
    def __init__(self, default, original):
        self.original = original
        self.default = default

    def __get__(self, oself, type=None):
        if type is not None:
            if oself is None:
                return self.default
        return getattr(oself, self.original, self.default)

    def __set__(self, oself, value):
        setattr(oself, self.original, value)

    def __delete__(self, oself):
        delattr(oself, self.original)


class Attribute(object):
    def __init__(self, doc=''):
        self.doc = doc

    def requiredSlots(self, modname, classname, attrname):
        self.name = attrname
        yield attrname

    def __get__(self, oself, type=None):
        assert oself is None, "%s: should be masked" % (self.name,)
        return self

_RAISE = object()
class SetOnce(Attribute):

    def __init__(self, doc='', default=_RAISE):
        Attribute.__init__(self)
        if default is _RAISE:
            self.default = ()
        else:
            self.default = (default,)

    def requiredSlots(self, modname, classname, attrname):
        self.name = attrname
        t = self.trueattr = ('_' + self.name)
        yield t

    def __set__(self, iself, value):
        if not hasattr(iself, self.trueattr):
            setattr(iself, self.trueattr, value)
        else:
            raise AttributeError('%s.%s may only be set once' % (
                    type(iself).__name__, self.name))

    def __get__(self, iself, type=None):
        if type is not None and iself is None:
            return self
        return getattr(iself, self.trueattr, *self.default)

class SchemaMetaMachine(_SlotMetaMachine):

    def determineSchema(meta, dictionary):
        attrs = dictionary['__attributes__'] = []
        name = dictionary['__name__']
        moduleName = dictionary['__module__']
        dictitems = dictionary.items()
        dictitems.sort()
        for k, v in dictitems:
            if isinstance(v, Attribute):
                attrs.append((k, v))
                for slot in v.requiredSlots(moduleName, name, k):
                    if slot == k:
                        del dictionary[k]
                    yield slot

    determineSchema = classmethod(determineSchema)


class _Strict(object):
    """
    I disallow all attributes from being set that do not have an explicit
    data descriptor.
    """

    def __setattr__(self, name, value):
        """
        Like PyObject_GenericSetAttr, but call descriptors only.
        """
        try:
            allowed = type(self).__dict__['_Strict__setattr__allowed']
        except KeyError:
            allowed = type(self)._Strict__setattr__allowed = {}
            for cls in type(self).__mro__:
                for attrName, slot in cls.__dict__.iteritems():
                    if attrName in allowed:
                        # It was found earlier in the mro, overriding
                        # whatever this is.  Ignore it and move on.
                        continue
                    setter = getattr(slot, '__set__', _NOSLOT)
                    if setter is not _NOSLOT:
                        # It is a data descriptor, so remember the setter
                        # for it in the cache.
                        allowed[attrName] = setter
                    else:
                        # It is something else, so remember None for it in
                        # the cache to indicate it cannot have its value
                        # set.
                        allowed[attrName] = None

        try:
            setter = allowed[name]
        except KeyError:
            pass
        else:
            if setter is not None:
                setter(self, value)
                return

        # It wasn't found in the setter cache or it was found to be None,
        # indicating a non-data descriptor which cannot be set.
        raise AttributeError(
            "%r can't set attribute %r" % (self.__class__.__name__, name))


class SchemaMachine(_Strict):
    __metaclass__ = SchemaMetaMachine

class SlotMachine(_Strict):
    __metaclass__ = _SlotMetaMachine



# -*- test-case-name: axiom.test.historic.test_subStoreStartupService1to2 -*-

from zope.interface import implements
from twisted.application.service import IService

from axiom.item import Item
from axiom.attributes import boolean
from axiom.substore import SubStore, SubStoreStartupService

from axiom.test.historic.stubloader import saveStub

class DummyService(Item):
    """
    Service which does nothing but mark itself as run, if it's ever run.  After
    the upgrader it should not be run.
    """
    typeName = 'substore_service_upgrade_stub_service'
    everStarted = boolean(default=False)
    implements(IService)

    name = property(lambda : "sucky-service")
    running = property(lambda : False)

    def setName(self, name):
        pass
    def setServiceParent(self, parent):
        pass
    def disownServiceParent(self):
        pass
    def startService(self):
        self.everStarted = True
    def stopService(self):
        pass
    def privilegedStartService(self):
        pass


def createDatabase(s):
    """
    Create a store which contains a substore-service-starter item powered up
    for IService, and a substore, which contains a service that should not be
    started after the upgrader runs.
    """
    ssi = SubStore.createNew(s, ["sub", "test"])
    ss = ssi.open()
    ds = DummyService(store=ss)
    ss.powerUp(ds, IService)
    ssss = SubStoreStartupService(store=s).installOn(s)

if __name__ == '__main__':
    saveStub(createDatabase, 7615)



# -*- test-case-name: axiom.test.test_upgrading.PathUpgrade.test_postCopy -*-

from axiom.attributes import path

from axiom.item import Item

from axiom.upgrade import registerAttributeCopyingUpgrader

class Path(Item):
    """
    Trivial Item class for testing upgrading.
    """
    schemaVersion = 2
    typeName = 'test_upgrade_path'
    thePath = path()

def fixPath(it):
    """
    An example postcopy function, for fixing up an item after its attributes
    have been copied.
    """
    it.thePath = it.thePath.child("foo")

registerAttributeCopyingUpgrader(Path, 1, 2, postCopy=fixPath)




import os
import sys
import shutil
import tarfile
import inspect

from twisted.trial import unittest
from twisted.application.service import IService

from axiom.store import Store

def saveStub(funcobj, revision):
    """
    Create a stub database and populate it using the given function.

    @param funcobj: A one-argument callable which will be invoked with an Axiom
    Store instance and should add to it the old state which will be used to
    test an upgrade.

    @param revision: An SVN revision of trunk at which it was possible it is
    possible for funcobj to create the necessary state.
    """
    # You may notice certain files don't pass the second argument.  They don't
    # work any more.  Please feel free to update them with the revision number
    # they were created at.
    filename = inspect.getfile(funcobj)
    dbfn = os.path.join(
        os.path.dirname(filename),
        os.path.basename(filename).split("stub_")[1].split('.py')[0]+'.axiom')

    s = Store(dbfn)
    s.transact(funcobj, s)

    s.close()
    tarball = tarfile.open(dbfn+'.tbz2', 'w:bz2')
    tarball.add(os.path.basename(dbfn))
    tarball.close()
    shutil.rmtree(dbfn)



class StubbedTest(unittest.TestCase):

    def openLegacyStore(self):
        """
        Extract the Store tarball associated with this test, open it, and return
        it.
        """
        temp = self.mktemp()
        f = sys.modules[self.__module__].__file__
        dfn = os.path.join(
            os.path.dirname(f),
            os.path.basename(f).split("test_")[1].split('.py')[0]+'.axiom')
        arcname = dfn + '.tbz2'
        tarball = tarfile.open(arcname, 'r:bz2')
        for member in tarball.getnames():
            tarball.extract(member, temp)
        return Store(os.path.join(temp, os.path.basename(dfn)))


    def setUp(self):
        """
        Prepare to test a stub by opening and then fully upgrading the legacy
        store.
        """
        self.store = self.openLegacyStore()
        self.service = IService(self.store)
        self.service.startService()
        return self.store.whenFullyUpgraded()


    def tearDown(self):
        return self.service.stopService()




from axiom.userbase import LoginSystem
from axiom.test.test_userbase import GarbageProtocolHandler
from axiom.test.historic.test_loginMethod1to2 import CREDENTIALS, GARBAGE_LEVEL

def createDatabase(s):
    ls = LoginSystem(store=s)
    ls.installOn(s)
    acc = ls.addAccount(*CREDENTIALS)
    ss = acc.avatars.open()
    gph = GarbageProtocolHandler(store=ss, garbage=GARBAGE_LEVEL)
    gph.installOn(ss)

from axiom.test.historic.stubloader import saveStub

if __name__ == '__main__':
    saveStub(createDatabase)
